# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

# ModelParameters imports
from modelparameters.logger import *
import modelparameters.commands as commands

# Base class for ModelParameters exceptions
class GotranException(RuntimeError):
    "Base class for ModelParameters exceptions"
    pass


set_default_exception(GotranException)
from modelparameters.utils import *

# from modelparameters.utils import (Range, Timer, list_types, clear_timings, tic,
#                              toc, is_iterable, add_iterable, camel_capitalize,
#                              tuplewrap, listwrap, check_arg, check_arginlist,
#                              check_kwarg, quote_join, deprecated, format_time,
#                              value_formatter, param2value)
# # gotran imports
from gotran.common.dicts import adict, odict
from gotran.common.disk import load, save, present_time_str
from gotran.common.options import parameters

# Set initial log level to INFO
set_log_level(INFO)

__all__ = [_name for _name in list(globals().keys()) if _name[0] != "_"]
# __all__ = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "Logger", "commands",
#            "Range", "Timer", "list_types", "clear_timings", "tic",
#            "toc", "is_iterable", "add_iterable", "camel_capitalize",
#            "tuplewrap", "listwrap", "check_arg", "check_arginlist",
#            "check_kwarg", "quote_join", "deprecated", "format_time",
#            "value_formatter", "param2value", "adict", "odict", "load", "save",
#            "present_time_str", "parameters"]



# Copyright (C) 2011-2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

# FIXME: This file is depricated...

__all__ = ["CellModel", "gccm"]

from gotran.common import *
from gotran.model.ode import ODE
from gotran.model.odeobjects import *

# from gotran.common import error, debug, check_arg, check_kwarg, check_arginlist

from modelparameters.parameters import *
import modelparameters.sympytools as sp_tools

# Holder for all cellmodels
_all_cellmodels = {}

# Holder for current CellModel
_current_cellmodel = None


def stimulation_protocal(stim_params):
    """
    Return stimulation protocal given a dictionary of stimuluation
    parameters
    """

    if stim_params:
        amplitude = get_parameter_list_from_string("amp", stim_params)[0]
        duration = get_parameter_list_from_string("dur", stim_params)[0]

        period = get_parameter_list_from_string("period", stim_params)
        frequency = get_parameter_list_from_string("frequency", stim_params)
        if period:
            period = period[0]
        if frequency:
            frequency = frequency[0]

        start = get_parameter_list_from_string("start", stim_params)[0]
        end = get_parameter_list_from_string("end", stim_params)[0]
    else:
        amplitude = Parameter("amplidude", ScalarParam(value=0))
        duration = Parameter("duration", ScalarParam(value=0))
        period = Parameter("period", ScalarParam(value=500, unit="ms"))
        start = Parameter("start", ScalarParam(value=0))
        end = Parameter("end", ScalarParam(value=30000, unit="ms"))
        frequency = None

    if not period and frequency:
        # Let the period be the reciprocal of the frequency
        period_param = 60.0 / frequency
        period = Parameter("period", period_param)
    if not frequency and period:
        frequency_param = 60.0 / period
        frequency = Parameter("frequency", frequency_param)

    class StimDict(dict):
        def set(self, key, value):
            self[key].update(value)
            if key == "period":
                self["frequency"].update(60.0 / value)
            elif key == "frequency":
                self["period"].update(60.0 / value)

    return StimDict(
        amplitude=amplitude,
        duration=duration,
        period=period,
        frequency=frequency,
        start=start,
        end=end,
    )


def get_parameter_list_from_string(string, lst, case_insesitive=True):
    """
    Return a list with parameters in the given list
    that has a name containing the given string

    Arguments
    ---------

    string : str
        The string that should be in the name
    lst : array with parameters
        List containing the candidates for parameters
        case_insesitive: bool
        If True, include parameter as long as the letters coincide
        and do not care about case

    Returns
    -------

    parlist : lst
        A sublist of lst containing the parameter that has a name
        containing the given string

    """
    if case_insesitive:
        return [p for p in lst if string.lower() in p.name.lower()]

    else:
        return [p for p in lst if string in p.name]


class CellModel(ODE):
    """
    Basic class for storing information of a cell model
    You can either initialize an empty CellModel similar to
    an ODE.
    (Maybe if we can create a library of th cellmodels
    then it the string matched that of the library we return
    the model in the library, e.g `cell = CellModel("Noble_1962")`)
    Or you can initialize the cell using an existing ODE
    If you have an ODE and want to have a cell, then save the
    ODE to an .ode file and load it using `load_cell` (in stead of
    `load_ode`)

    Example
    -------

    .. code-block::python

        cell =  CellModel("MyCell")


    Example
    -------

    .. code-block::python

        # Save current ODE
        ode.save(filename)

        # Load cell
        from loadmodel import load_cell
        cell = load_cell(filename)

    """

    def __new__(cls, *args, **kwargs):
        """
        Create a CellModel instance.
        """

        arg = args[0]
        if arg in _all_cellmodels:
            return _all_cellmodels[arg]

        return object.__new__(cls)

    def __init__(self, name_, ns=None):
        """
        Initialize a CellModel

        """

        check_arg(name_, str)
        name = name_

        super(CellModel, self).__init__(name, ns)
        # Do not reinitialized object if it already excists
        if name in _all_cellmodels:
            return

        # Initialize attributes
        self.name = name

        self._initialize_cell_model()

        # Store instance for future lookups
        _all_cellmodels[name] = self

    def _initialize_cell_model(self):
        # Set current CellModel
        global _current_cellmodel
        _current_cellmodel = self
        # Perhaps more here later

    @property
    def parameter_symbols(self):
        return [s.name for s in self.parameters]

    @property
    def component_names(self):
        return [s.name for s in self.components]

    def parameter_values(self):
        return [s.value for s in self.parameters]

    @property
    def state_symbols(self):
        return [s.name for s in self.states]

    def state_values(self):
        return [s.value for s in self.states]

    def intermediate_unit(self, name, unit_type="si", return_factor=False):
        """
        Get unit of intermediate expression
        Note that we neglect units within a funtion like
        exponential and logaritm.

        Arguments
        ---------
        name : str
            Name of intermediate
        unit_type : str
            Type of unit, options `si`, `base` or `original`

        Returns
        -------
        unit : str
            Unit of the expression for the intermediate

        """

        def get_intermediate_unit(name, unit_type):
            check_arginlist(name, self.intermediate_symbols)
            intermediate = self.intermediates[self.intermediate_symbols.index(name)]

            factor = 1.0
            # Extract expression
            expr = intermediate.expr.copy()

            if isinstance(expr, sp.Piecewise):
                expr = expr.args[0][0]

            expr = expr.replace(sp.log, lambda t: 1)
            expr = expr.replace(sp.exp, lambda t: 1)
            expr = expr.replace(sp.zoo, lambda: 1)

            unit_dep_map = {}

            for dep in sp_tools.symbols_from_expr(expr):

                dep_str = str(dep).rsplit("(")[0]

                if dep_str in self.parameter_symbols:
                    p = self.get_parameter(str(dep))
                    unit = p.unit

                elif dep_str in self.state_symbols:
                    p = self.get_state(dep_str)
                    unit = p.unit

                elif dep_str in self.intermediate_symbols:
                    unit, factor_ = get_intermediate_unit(dep_str, "original")
                    # factor *= factor_

                else:
                    # Parmaterer not found (most likely the time variable)
                    continue

                expr1 = expr.subs(dep, unit)
                if expr1 == 0:
                    # We got some cancellation because of the previous substitutioin

                    # Let us add a dummy number
                    expr = expr.subs(dep, "*".join(["2.0", unit]))
                else:
                    expr = expr1

                unit_dep_map[dep] = unit

            # Substitute again
            for k, v in list(unit_dep_map.items()):
                for k1, v1 in list(unit_dep_map.items()):
                    k = k.subs(k1, v1)
                expr = expr.subs(k, v)

            # Fix fractions and remove possible numbers
            unit_exprs = []

            def add_unit(k, v):
                if not k.is_Number and v.as_numer_denom()[1] == 1:

                    exp = "**{}".format(str(v)) if v != 1 else ""
                    unit_ = str(k)

                    unit_exprs.append(unit_ + exp)

            for k, v in list(expr.as_powers_dict().items()):

                # If term consist of multiple term, choose one of them
                if k.is_Add:
                    k = k.as_coeff_add()[1][0]

                    for k1, v1 in list(k.as_powers_dict().items()):
                        add_unit(k1, v1)

                else:
                    add_unit(k, v)

            # Join by multiplication
            unit_expr = "*".join(unit_exprs)

            # Check if this is a sum and use only one term
            unit_expr = unit_expr.split(" + ")[0].split(" - ")[0]

            # Strip away any numbers but collect the factor
            subunits = "^".join(unit_expr.split("**")).split("*")
            new_subunits = []

            def isfloat(el):
                try:
                    float(el)
                except:
                    return False
                else:
                    return True

            for u in subunits:
                if not isfloat(u):
                    new_subunits.append(u)

            # Join new expression
            unit_expr = "**".join("*".join(new_subunits).split("^"))

            if unit_expr == "":
                unit_expr = "1"

            unit = Unit(unit_expr)

            if unit_type == "si":
                factor = unit.si_unit_factor
                retunit = unit.si_unit
            elif unit_type == "base":
                factor = unit.factor
                retunit = unit.base_unit
            elif unit_type == "original":
                factor = 1.0
                retunit = unit.unit

            return retunit, factor

        unit_, factor_ = get_intermediate_unit(name, unit_type)

        if return_factor:
            return unit_, factor_
        return unit_

    def set_residual_current(self, t, current):
        """
        Set rediual current

        Arguments
        ---------

        t : array
            List of times
        current : array
            List with residual current

        """
        from scipy.interpolate import UnivariateSpline

        self._residual_current = UnivariateSpline(t, current, s=0)

    def residual_current(self, t):

        if not hasattr(self, "_residual_current"):
            return np.zeros_like(t)

        return self._residual_current(t)

    @property
    def intermediate_symbols(self):
        return [i.name for i in self.intermediates]

    @property
    def stimulation_parameters(self):
        return get_parameter_list_from_string("stim", self.parameters)

    @property
    def stimulation_protocol(self):

        stim_params = self.stimulation_parameters

        return stimulation_protocal(self.stimulation_parameters)

    @property
    def currents(self):
        """
        Return a list of the currents used in the
        computation of the membrane potential.
        Note that intermediate currents (not involved in
        th computation of the membrane potential)
        are not retured
        """

        dV_dt = get_parameter_list_from_string("dV_dt", self.state_expressions)[0]
        currents = [str(type(c)) for c in dV_dt.dependent]
        return currents

    def get_beattime(self, dt=1.0, extra=60.0):
        """
        Return number timepoints for one beat

        Arguments
        ---------
        dt : float
            Time increment in the same time unit
            as the model
        extra : float
            Add some extra time (in ms) to prolonge the
            the time. Default 60 ms

        """

        # Get stimulation parameters
        stim_params = self.stimulation_protocol

        # Get duration of one beat
        if stim_params["period"]:
            period = stim_params["period"].value
        else:
            period = 60.0 / stim_params["frequency"].value

        # Get duration of simulus
        duration = stim_params["duration"].value

        factor = 1e-3 if stim_params["duration"].param.unit == "s" else 1.0
        # Include additional 60 ms before stimulation
        extra_ = factor * extra
        beattime = int((period + duration + extra_) / float(dt)) + 1

        return beattime

    def get_time_steps(self, nbeats=1, t1=None, dt=1.0, t0=0.0):
        """
        Get list with time steps given the number
        of beats and time increment

        Arguments
        ---------
        nbeats : int
            Nuber of beats (Default:1)
        dt : float
            Time increament between two time steps in ms. (Default:1.0)
            Note that you can provide `dt` as a Parameter with unit different
            from ms.
        t1 : float
            End time. If not provided then end time will determined from
            the number of beats
        t0 : float
            Start time (Default: 0.0)

        """
        if isinstance(dt, scalars):
            dt = ScalarParam(dt, unit="ms")
        # Get stimulation prototocal
        stim_params = self.stimulation_protocol
        # We let the period of the frequency
        # define the lenght of each beat
        period = stim_params["period"]

        if t1 is None:
            # Use the stimultation protocol to determine the end time
            t1 = t0 + nbeats * period

        # Estimate number of steps
        nsteps = t1 / dt + 1
        tsteps = np.linspace(
            param2value(t0), param2value(t1), param2value(nsteps.value)
        )
        return tsteps

    def simulate(self, **kwargs):
        """
        Simulate the ODE to :math:`t_{\mathrm{end}}`
        with the given number points

        Arguments
        ---------

        t_end : scalar
            The end time
        nbeats : scalar
            Number of beats based on stimulus protocol
        npts : int
            Number of communication points used in the solver

        """

        # Parameters for time
        t_end = kwargs.pop("t_end", None)
        dt = kwargs.pop("dt", 0.1)

        # Number of beats
        nbeats = kwargs.pop("nbeats", 1)

        # Number of points (only assimulo)
        npts = kwargs.pop("npts", None)

        # Solver backend
        backend = kwargs.pop("backend", "goss")
        # Solver meothd
        method = kwargs.pop("solver_method", "RKF32")
        # Residual current
        update_residual_current = kwargs.pop("residual_current", False)
        # Return only the final beat (if simulating multiple beats)
        return_final_beat = kwargs.pop("return_final_beat", False)
        # Return monitored values
        return_monitored = kwargs.pop("return_monitored", False)

        # Get stimulation prototocal
        stim_params = self.stimulation_protocol

        # We let the period of the frequency
        # define the lenght of each beat
        if stim_params["period"]:
            period = stim_params["period"].value
        else:
            period = 60.0 / stim_params["frequency"].value

        if t_end is None:
            # Use the stimultation protocol to determine the end time
            t0 = stim_params["start"]
            t_end = t0 + nbeats * period

        # Estimate number of steps
        nsteps = int(t_end / float(dt))

        # Find start and end indices
        if return_final_beat:
            start_idx = int(
                nsteps * (1 - (period + stim_params["start"].value) / float(t_end)) - 1
            )
            end_idx = int(nsteps)
        else:
            start_idx = 0
            end_idx = nsteps

        # Set the end value for stimulation
        stim_params["end"].param.setvalue(t_end, False)

        if backend == "goss":

            import goss

            msg = "Solver method has to be one of " + "{}, got {}".format(
                goss.goss_solvers, method
            )
            assert method in goss.goss_solvers, msg

            monitored_symbols = self.intermediate_symbols if return_monitored else None

            module = goss.jit(self, monitored=monitored_symbols)

            solver = getattr(goss, method)(module)
            x0 = module.init_state_values()
            t = 0.0

            ys = np.zeros((nsteps, module.num_states()))
            if return_monitored:
                monitored = np.zeros((nsteps, module.num_monitored()))
                monitor = np.zeros(module.num_monitored())
            ts = np.zeros(nsteps)

            for step in range(nsteps):

                if update_residual_current:

                    module.set_parameter("i_res_amp", float(self.residual_current(t)))
                    # solver = getattr(goss, method)(module)
                    # print float(self.residual_current(t))

                if return_monitored:
                    module.eval_monitored(x0, t, monitor)
                    monitored[step, :] = monitor

                # print t
                solver.forward(x0, t, dt)

                ys[step, :] = x0
                ts[step] = t
                t += dt

            ret = [ts, ys]

            if return_monitored:
                ret.append(monitored)

        else:
            from gotran.codegeneration.codegenerators import PythonCodeGenerator
            from gotran.common.options import parameters
            import imp

            params = parameters.generation.copy()
            params.functions.rhs.function_name = "__call__"
            params.code.default_arguments = "tsp"
            params.class_code = 1

            monitored = [
                expr.name for expr in self.intermediates + self.state_expressions
            ]
            gen = PythonCodeGenerator(params)

            name = self.name
            self.rename("ODESim")
            code = gen.class_code(self, monitored)
            self.rename(name)

            module = imp.new_module("simulation")
            exec(code, module.__dict__)

            ode = module.ODESim()
            from cellmodels.odesolver import ODESolver

            states = ode.init_state_values()

            options = ODESolver.list_solver_options(method)
            atol = kwargs.pop("atol", None)

            if atol:
                options["atol"] = atol * np.ones(len(states))

            for k, v in kwargs.items():
                if v and k in options:
                    options[k] = v

            solver = ODESolver(ode, states, method=method, **options)

            t_, y_ = solver.solve(t_end, nbeats)

            ret = [t_, y_]

        if return_final_beat:

            ret[0] = ret[0][start_idx:end_idx]
            ret[1] = ret[1][start_idx:end_idx]

        return ret

    def get_component(self, name):
        """
        Get the component with the given name

        Arguments
        ---------

        name : str
            Name of the component

        Returns
        -------
        par : gotran.ODECompoenent
            The component

        """

        # Check if parameter exist
        check_arginlist(name, self.component_names)
        # Get the index
        idx = self.component_names.index(name)
        # Return the component
        return self.components[idx]

    def get_parameter(self, name):
        """
        Get the parameter with the given name

        Arguments
        ---------

        name : str
            Name of the parameter

        Returns
        -------
        par : gotran.Parameter
            The parameter

        """
        # Check if parameter exist
        check_arginlist(name, self.parameter_symbols)
        # Get the index
        idx = self.parameter_symbols.index(name)
        # Return the parameter
        return self.parameters[idx]

    def get_state(self, name):
        """
        Get the state with the given name

        Arguments
        ---------

        name : str
            Name of the state variable

        Returns
        -------
        par : gotran.Parameter
            The state

        """
        # Check if parameter exist
        check_arginlist(name, self.state_symbols)
        # Get the index
        idx = self.state_symbols.index(name)
        # Return the parameter
        return self.states[idx]

    def get_intermediate(self, name):
        """
        Get the intermediate with the given name

        Arguments
        ---------

        name : str
            Name of the intermediate

        Returns
        -------
        par : gotran.Parameter
            The parameter

        """
        # Check if parameter exist
        check_arginlist(name, self.intermediate_symbols)
        # Get the index
        idx = self.intermediate_symbols.index(name)
        # Return the parameter
        return self.intermediates[idx]

    def get_parameter(self, name):
        """
        Get the parameter with the given name

        Arguments
        ---------

        name : str
            Name of the parameter

        Returns
        -------
        par : gotran.Parameter
            The parameter

        """
        # Check if parameter exist
        check_arginlist(name, self.parameter_symbols)
        # Get the index
        idx = self.parameter_symbols.index(name)
        # Return the parameter
        return self.parameters[idx]

    def set_parameter(self, name, value):
        """
        Set the parameter in the model to a
        given value


        Arguments
        ---------

        name : str
            Name of the parmaeter
        value : scalar, gotran.ScalarParam
            The new value of the parameter. Note that
            if the parameter is of type `ScalarParam`
            while the provided value is a scalar then the value
            will be updated while keeping the unit

        """
        check_arg(value, scalars + (ScalarParam,), 1, Parameter)

        par = self.get_parameter(name)
        if isinstance(value, ScalarParam):
            par._param = value
        else:
            par.value = value

    def update_parameter(self, name, value=None, factor=1.0):

        if value is None:
            par = self.get_parameter(name)
            par._param.value *= factor
        else:
            self.set_parameter(name, value)


# Construct a default CellModel
# _current_cellmodel = CellModel("Default")


def gccm():
    """
    Return the current CellModel
    """
    assert isinstance(_current_cellmodel, CellModel)
    return _current_cellmodel



#!/usr/bin/env python3

from gotran.codegeneration.latexcodegenerator import (
    LatexCodeGenerator,
    _default_latex_params,
)
from gotran.model.loadmodel import load_ode
from gotran import info


def gotran2latex(filename, params):
    """
    Create LaTeX code from a gotran model
    """

    if not params.sympy_contraction:
        # A hack to avoid sympy contractions
        import gotran.codegeneration.avoidsympycontractions

    # Load Gotran model
    ode = load_ode(filename)

    # Check for completeness
    # TODO: Should raise a more descriptive exception?
    if not ode.is_complete:
        raise Exception("Incomplete ODE")

    # Create a gotran -> LaTeX document generator
    gen = LatexCodeGenerator(ode, params)

    info("")
    info("Generating LaTeX files for the {0} ode...".format(ode.name))
    with open(gen.output_file, "w") as f:
        f.write(gen.generate())
    info("  done.")


def main():
    import os
    import sys
    from modelparameters.parameterdict import ParameterDict, Param

    params = _default_latex_params()
    params = ParameterDict(
        sympy_contraction=Param(
            True,
            description="If True sympy contraction"
            " will be used, turning (V-3)/2 into V/2-3/2",
        ),
        **params
    )
    params.parse_args(usage="usage: %prog FILE [options]")

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    filename = sys.argv[1]
    gotran2latex(filename, params)


if __name__ == "__main__":
    main()



# Copyright (C) 2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

# Import gotran modules
from . import codegenerators
from . import compilemodule
from . import latexcodegenerator
from . import codecomponent
from . import algorithmcomponents
from . import oderepresentation
from . import solvercomponents

# Import classes and routines from gotran modules
from .codecomponent import CodeComponent
from .algorithmcomponents import (
    JacobianComponent,
    JacobianActionComponent,
    FactorizedJacobianComponent,
    ForwardBackwardSubstitutionComponent,
    LinearizedDerivativeComponent,
    CommonSubExpressionODE,
    componentwise_derivative,
    linearized_derivatives,
    jacobian_expressions,
    jacobian_action_expressions,
    factorized_jacobian_expressions,
    forward_backward_subst_expressions,
    diagonal_jacobian_expressions,
    rhs_expressions,
    diagonal_jacobian_action_expressions,
    monitored_expressions,
)
from .solvercomponents import (
    ExplicitEuler,
    explicit_euler_solver,
    RushLarsen,
    rush_larsen_solver,
    GeneralizedRushLarsen,
    generalized_rush_larsen_solver,
    HybridGeneralizedRushLarsen,
    hybrid_generalized_rush_larsen_solver,
    SimplifiedImplicitEuler,
    simplified_implicit_euler_solver,
    get_solver_fn,
)
from .codegenerators import (
    PythonCodeGenerator,
    CCodeGenerator,
    CppCodeGenerator,
    MatlabCodeGenerator,
    class_name,
    CUDACodeGenerator,
    JuliaCodeGenerator,
)
from .latexcodegenerator import LatexCodeGenerator, _default_latex_params
from .compilemodule import compile_module

# from .oderepresentation import *


# Assign the __all__ attribute
# __all__ = [name for name in list(globals().keys()) if name[:1] != "_"]
__all__ = [
    "codecomponent",
    "CodeComponent",
    "algorithmcomponents",
    "JacobianComponent",
    "JacobianActionComponent",
    "FactorizedJacobianComponent",
    "ForwardBackwardSubstitutionComponent",
    "LinearizedDerivativeComponent",
    "CommonSubExpressionODE",
    "componentwise_derivative",
    "linearized_derivatives",
    "jacobian_expressions",
    "jacobian_action_expressions",
    "factorized_jacobian_expressions",
    "forward_backward_subst_expressions",
    "diagonal_jacobian_expressions",
    "rhs_expressions",
    "diagonal_jacobian_action_expressions",
    "monitored_expressions",
    "solvercomponents",
    "JacobianComponent",
    "JacobianActionComponent",
    "FactorizedJacobianComponent",
    "ForwardBackwardSubstitutionComponent",
    "LinearizedDerivativeComponent",
    "CommonSubExpressionODE",
    "componentwise_derivative",
    "linearized_derivatives",
    "jacobian_expressions",
    "jacobian_action_expressions",
    "factorized_jacobian_expressions",
    "forward_backward_subst_expressions",
    "diagonal_jacobian_expressions",
    "rhs_expressions",
    "diagonal_jacobian_action_expressions",
    "monitored_expressions",
    "codegenerators",
    "PythonCodeGenerator",
    "CCodeGenerator",
    "CppCodeGenerator",
    "MatlabCodeGenerator",
    "class_name",
    "CUDACodeGenerator",
    "JuliaCodeGenerator",
    "latexcodegenerator",
    "LatexCodeGenerator",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import division

__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2013-03-13 -- 2015-06-24"
__copyright__ = "Copyright (C) 2013 " + __author__
__license__ = "GNU LGPL Version 3.0 or later"
from modelparameters.codegeneration import latex

try:
    from scipy.optimize import root
except:
    root = None
from itertools import cycle
import matplotlib.pyplot as plt
import numpy as np
import instant
from gotran.model.loadmodel import load_ode
from gotran.model.utils import DERIVATIVE_EXPRESSION, special_expression
from gotran.codegeneration.compilemodule import compile_module
from gotran.common.options import parameters
from gotran.common import error, warning
from modelparameters.parameterdict import *


def gotranrun(filename, params):

    # Copy of default parameters
    generation = parameters.generation.copy()

    # Set body parameters
    generation.code.body.representation = params.code.body_repr
    generation.code.body.use_cse = params.code.use_cse
    generation.code.body.optimize_exprs = params.code.optimize_exprs

    # Set what code we are going to generate and not
    for what_not in [
        "componentwise_rhs_evaluation",
        "forward_backward_subst",
        "linearized_rhs_evaluation",
        "lu_factorization",
        "jacobian",
    ]:
        generation.functions[what_not].generate = False

    # Always generate code for monitored expressions
    generation.functions.monitored.generate = True

    # If scipy is used to solve we generate RHS and potentially a jacobian
    if params.solver == "scipy":
        generation.functions.rhs.generate = True
        generation.functions.jacobian.generate = params.code.generate_jacobian
    else:
        generation.solvers[params.solver].generate = True

    # Compile executeable code from gotran ode
    model_arguments = params.model_arguments
    if len(model_arguments) == 1 and model_arguments[0] == "":
        model_arguments = []

    if len(model_arguments) % 2 != 0:
        error("Expected an even number of values for 'model_arguments'")

    arguments = dict()
    for arg_name, arg_value in [
        (model_arguments[i * 2], model_arguments[i * 2 + 1])
        for i in range(int(len(model_arguments) / 2))
    ]:

        arguments[arg_name] = arg_value

    ode = load_ode(filename, **arguments)

    # Check for DAE
    if ode.is_dae:
        error(
            "Can only integrate pure ODEs. {0} includes algebraic states "
            "and is hence a DAE.".format(ode.name)
        )

    # Get monitored and plot states
    plot_states = params.plot_y

    # Get x_values
    x_name = params.plot_x

    state_names = [state.name for state in ode.full_states]
    monitored_plot = [
        plot_states.pop(plot_states.index(name))
        for name in plot_states[:]
        if name not in state_names
    ]

    monitored = []
    all_monitored_names = []
    for expr in sorted(ode.intermediates + ode.state_expressions):
        if expr.name not in monitored:
            monitored.append(expr.name)
        all_monitored_names.append(expr.name)

    # Check valid monitored plot
    for mp in monitored_plot:
        if mp not in monitored:
            error("{} is not a state or intermediate in this ODE".format(mp))

    # Check x_name
    if x_name not in ["time"] + monitored + state_names:
        error(
            "Expected plot_x to be either 'time' or one of the plotable "
            "variables, got {}".format(x_name)
        )

    # Logic if x_name is not 'time' as we then need to add the name to
    # either plot_states or monitored_plot
    if x_name != "time":
        if x_name in state_names:
            plot_states.append(x_name)
        else:
            monitored_plot.append(x_name)

    module = compile_module(ode, params.code.language, monitored, generation)

    parameter_values = params.parameters
    init_conditions = params.init_conditions

    if len(parameter_values) == 1 and parameter_values[0] == "":
        parameter_values = []

    if len(init_conditions) == 1 and init_conditions[0] == "":
        init_conditions = []

    if len(parameter_values) % 2 != 0:
        error("Expected an even number of values for 'parameters'")

    if len(init_conditions) % 2 != 0:
        error("Expected an even number of values for 'initial_conditions'")

    user_params = dict()
    for param_name, param_value in [
        (parameter_values[i * 2], parameter_values[i * 2 + 1])
        for i in range(int(len(parameter_values) / 2))
    ]:

        user_params[param_name] = float(param_value)

    user_ic = dict()
    for state_name, state_value in [
        (init_conditions[i * 2], init_conditions[i * 2 + 1])
        for i in range(int(len(init_conditions) / 2))
    ]:

        user_ic[state_name] = float(state_value)

    # Use scipy to integrate model
    t0 = 0.0
    t1 = params.tstop
    dt = params.dt

    rhs = module.rhs
    jac = module.compute_jacobian if params.code.generate_jacobian else None
    y0 = module.init_state_values(**user_ic)
    model_params = module.init_parameter_values(**user_params)

    # Check for steady state solve
    if params.steady_state.solve and root:
        result = root(
            rhs,
            y0,
            args=(0.0, model_params),
            jac=jac,
            method=params.steady_state.method,
            tol=params.steady_state.tol,
        )

        if result.success:
            y0 = result.x
            print(
                "Found stead state:",
                ", ".join(
                    "{0}: {1:e}".format(state.name, value)
                    for value, state in zip(y0, ode.full_states)
                ),
            )
        else:
            warning(result.message)

    tsteps = np.linspace(t0, t1, int(t1 / dt) + 1)

    # What solver should we use
    if params.solver == "scipy":
        try:
            from scipy.integrate import odeint
        except Exception as e:
            error("Problem importing scipy.integrate.odeint. {}".format(e))
        results = odeint(rhs, y0, tsteps, Dfun=jac, args=(model_params,))

    else:

        # Get generated forward method
        forward = getattr(module, "forward_" + params.solver)

        results = [y0]
        states = y0.copy()

        # Integrate solution using generated forward method
        for ind, t in enumerate(tsteps[:-1]):

            # Step solver
            forward(states, t, dt, model_params)
            results.append(states.copy())

    # Plot results
    if not (plot_states or monitored or params.save_results):
        return

    # allocate memory for saving results
    if params.save_results:
        save_results = np.zeros(
            (len(results), 1 + len(state_names) + len(all_monitored_names))
        )
        all_monitor_inds = np.array(
            [monitored.index(monitor) for monitor in all_monitored_names], dtype=int
        )
        all_results_header = ", ".join(["time"] + state_names + all_monitored_names)

    plot_inds = [module.state_indices(state) for state in plot_states]

    monitored_values = [[] for monitor in monitored_plot]
    monitor_inds = np.array(
        [monitored.index(monitor) for monitor in monitored_plot], dtype=int
    )
    monitored_get_values = np.zeros(len(monitored), dtype=np.float_)

    # Allocate memory
    plot_values = np.zeros((len(plot_states) + len(monitored_plot), len(results)))

    for ind, (time, res) in enumerate(zip(tsteps, results)):

        if plot_states:
            plot_values[: len(plot_states), ind] = res[plot_inds]
        if monitored_plot or params.save_results:
            module.monitor(res, time, model_params, monitored_get_values)
        if monitored_plot:
            plot_values[len(plot_states) :, ind] = monitored_get_values[monitor_inds]
        if params.save_results:
            save_results[ind, 0] = time
            save_results[ind, 1 : len(state_names) + 1] = res
            save_results[ind, len(state_names) + 1 :] = monitored_get_values[
                all_monitor_inds
            ]

    # Save data
    if params.save_results:
        np.savetxt(
            "{}.csv".format(params.basename),
            save_results,
            header=all_results_header,
            delimiter=", ",
        )

    # Plot data

    if not (plot_states + monitored_plot):
        return

    # Fixes for derivatives
    monitored_plot_updated = []
    for monitor in monitored_plot:
        expr, what = special_expression(monitor, ode)
        if what == DERIVATIVE_EXPRESSION:
            var, dep = expr.groups()
            if var in ode.present_ode_objects and dep in ode.present_ode_objects:
                monitored_plot_updated.append("d{0}/d{1}".format(var, dep))
            else:
                monitored_plot_updated.append(monitor)
        else:
            monitored_plot_updated.append(monitor)

    plot_items = plot_states + monitored_plot
    if x_name != "time":
        x_values = plot_values[plot_items.index(x_name)]
    else:
        x_values = tsteps

    plt.rcParams["lines.linewidth"] = 2
    # line_styles = cycle([c+s for s in ["-", "--", "-.", ":"]
    # for c in plt.rcParams["axes.color_cycle"]])
    line_styles = cycle(
        [
            c + s
            for s in ["-", "--", "-.", ":"]
            for c in ["b", "g", "r", "c", "m", "y", "k"]
        ]
    )

    plotted_items = 0
    for what, values in zip(plot_items, plot_values):
        if what == x_name:
            continue
        plotted_items += 1

        plt.plot(x_values, values, next(line_styles))

    if plotted_items > 1:
        plt.legend(["$\\mathrm{{{0}}}$".format(latex(value)) for value in plot_items])
    elif plot_items:
        plt.ylabel("$\\mathrm{{{0}}}$".format(latex(plot_items[0])))

    plt.xlabel("$\\mathrm{{{0}}}$".format(latex(x_name)))
    plt.title(ode.name.replace("_", "\\_"))
    plt.show()


def main():
    import sys, os

    body_params = parameters.generation.code.body.copy()

    code_params = ParameterDict(
        language=OptionParam("C", ["Python", "C"]),
        body_repr=dict.__getitem__(body_params, "representation"),
        use_cse=dict.__getitem__(body_params, "use_cse"),
        optimize_exprs=dict.__getitem__(body_params, "optimize_exprs"),
        generate_jacobian=Param(
            False, description="Generate and use analytic " "jacobian when integrating."
        ),
    )

    steady_state = ParameterDict(
        solve=Param(
            False,
            description="If true scipy.optimize.root is used "
            "to find a steady state for a given parameters.",
        ),
        method=OptionParam(
            "hybr",
            [
                "hybr",
                "lm",
                "broyden1",
                "broyden2",
                "anderson",
                "linearmixing",
                "diagbroyden",
                "excitingmixing",
                "krylov",
            ],
        ),
        tol=ScalarParam(1e-5, description="Tolerance for root finding algorithm."),
    )

    solver = OptionParam(
        "scipy",
        ["scipy"] + list(parameters.generation.solvers.keys()),
        description="The solver that will be used to " "integrate the ODE.",
    )

    params = ParameterDict(
        solver=solver,
        steady_state=steady_state,
        parameters=Param([""], description="Set parameter of model"),
        init_conditions=Param([""], description="Set initial condition of model"),
        tstop=ScalarParam(100.0, gt=0, description="Time for stopping simulation"),
        dt=ScalarParam(0.1, gt=0, description="Timestep for plotting."),
        plot_y=Param(["V"], description="States or monitored to plot on the y axis."),
        plot_x=Param(
            "time",
            description="Values used for the x axis. Can be time "
            "and any valid plot_y variable.",
        ),
        model_arguments=Param([""], description="Set model arguments of the model"),
        code=code_params,
        save_results=Param(
            False,
            description="If True the results will be " "saved to a 'results.csv' file.",
        ),
        basename=Param(
            "results",
            description="The basename of the results "
            "file if the 'save_results' options is True.",
        ),
    )

    params.parse_args(usage="usage: %prog FILE [options]")

    if len(sys.argv) < 2:
        error("Expected a single gotran file argument.")

    if not os.path.isfile(sys.argv[1]):
        error("Expected the argument to be a file.", exception=IOError)

    file_name = sys.argv[1]
    gotranrun(file_name, params)


if __name__ == "__main__":
    main()



from . import cellml2gotran
from . import gotran2c
from . import gotran2cpp
from . import gotran2cuda
from . import gotran2dolfin
from . import gotran2julia
from . import gotran2latex
from . import gotran2matlab
from . import gotran2opencl
from . import gotran2py
from . import gotranexport
from . import gotranprobe
from . import gotranrun



# Copyright (C) 2013-2014 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ODEComponent"]

# System imports
from collections import OrderedDict, defaultdict
import weakref
from functools import partial

from sympy.core.function import AppliedUndef

# ModelParameters imports
from modelparameters.sympytools import sp, symbols_from_expr
from modelparameters.utils import Timer
from modelparameters.codegeneration import sympycode, _all_keywords

# Local imports
from gotran.common import error, debug, check_arg, check_kwarg, scalars
from gotran.model.odeobjects import *
from gotran.model.expressions import *
from gotran.model.utils import *


class ODEComponent(ODEObject):
    """
    Base class for all ODE components.
    """

    def __init__(self, name, parent):
        """
        Create an ODEComponent

        Arguments
        ---------
        name : str
            The name of the component. This str serves as the unique
            identifier of the Component.
        parent : gotran.ODEComponent
            The parent component of this ODEComponent
        """

        # Turn off magic attributes (see __setattr__ method) during
        # construction
        self._allow_magic_attributes = False

        check_arg(name, str, 0, ODEComponent)
        check_arg(parent, ODEComponent, 1, ODEComponent)

        # Call super class
        super(ODEComponent, self).__init__(name)

        # Store parent component
        self._parent = weakref.ref(parent)

        # Store ODEComponent children
        self.children = OrderedDict()

        # Store ODEObjects of this component
        self.ode_objects = ODEObjectList()

        # Storage of rates
        self.rates = RateDict(self)

        # Store all state expressions
        self._local_state_expressions = dict()

        # Collect all comments
        self._local_comments = []

        # Flag to check if component is finalized
        self._is_finalized = False
        self._is_finalizing = False

        # Turn on magic attributes (see __setattr__ method)
        self._allow_magic_attributes = True

    @property
    def parent(self):
        """
        Return the the parent if it is still alive. Otherwise it will return None
        """
        p = self._parent()
        if p is None:
            error("Cannot access parent ODEComponent. It is already " "destroyed.")
        return p

    def get_object(self, name, reversed=True, return_component=False):

        # First check self
        obj = self.ode_objects.get(name)

        if obj:
            return self, obj

        # Then iterate all objects in ode
        comp, obj = None, None

        # If a name is registered
        if name in self.root.ns:

            for obj in iter_objects(self, True, False, reversed):
                if isinstance(obj, ODEComponent):
                    comp = obj
                elif obj.name == name:
                    break
            else:
                comp, obj = None, None

        return comp, obj if return_component else obj

    @property
    def t(self):
        """
        Return the time symbol
        """
        return self.root._time.sym

    @property
    def time(self):
        """
        Return the time
        """
        return self.root._time

    def add_comment(self, comment, dependent=None):
        """
        Add a comment to the ODE component

        Arguments
        ---------
        comment : str
            The comment
        dependent : gotran.ODEObject
            If given the count of this comment will follow as a
            fractional count based on the count of the dependent object
        """
        comment = Comment(comment, dependent)
        self.ode_objects.append(comment)
        self._local_comments.append(comment)

    def add_state(self, name, init):
        """
        Add a state to the component

        Arguments
        ---------
        name : str
            The name of the state variable
        init : scalar, modelparameters.ScalarParam
            The initial value of the state
        """
        timer = Timer("Add states")

        # Create state
        state = State(name, init, self.time)

        self._register_component_object(state)

        # Return the sympy version of the state
        return state.sym

    def add_states(self, *args, **kwargs):
        """
        Add a number of states to the current ODEComponent

        Arguments
        ---------
        args : list of tuples
            A list of tuples with states and init values. Use this to
            set states if you need them ordered.
        kwargs : dict
            A dict with states
        """

        states = list(args) + sorted(kwargs.items())

        if len(states) == 0:
            error("Expected at least one state")

        for arg in states:
            if not isinstance(arg, tuple) or len(arg) != 2:
                error(
                    "excpected tuple with lenght 2 with state name (str) "
                    "and init values as the args argument."
                )
            state_name, init = arg

            # Add the states
            self.add_state(state_name, init)

    def add_parameter(self, name, init):
        """
        Add a parameter to the component

        Arguments
        ---------
        name : str
            The name of the parameter
        init : scalar, modelparameters.ScalarParam
            The initial value of the parameter
        """
        timer = Timer("Add parameters")

        param = Parameter(name, init)

        self._register_component_object(param)

        # Return the sympy version of the state
        return param.sym

    def add_parameters(self, *args, **kwargs):
        """
        Add a number of parameters to the current ODEComponent

        Arguments
        ---------
        args : list of tuples
            A list of tuples with parameters and init values. Use this to
            set parameters if you need them ordered.
        kwargs : dict
            A dict with parameters
        """

        params = list(args) + sorted(kwargs.items())

        if len(params) == 0:
            error("expected at least one parameter")

        for arg in params:
            if not isinstance(arg, tuple) or len(arg) != 2:
                error(
                    "excpected tuple with lenght 2 with parameter name (str) "
                    "and init values as the args argument."
                )
            parameter_name, value = arg

            # Add the parameters
            self.add_parameter(parameter_name, value)

    def add_component(self, name):
        """
        Add a sub ODEComponent
        """

        comp = ODEComponent(name, self)

        if name in self.root.all_components:
            error("A component with the name '{0}' already excists.".format(name))

        self.children[comp.name] = comp
        self.root.all_components[comp.name] = comp
        self.root._present_component = comp.name

        return comp

    def add_solve_state(self, state, expr, dependent=None, **solve_flags):
        """
        Add a solve state expression which tries to find a solution to
        a state by solving an algebraic expression

        Arguments
        ---------
        state : gotran.State, AppliedUndef
            The State that is solved
        expr : sympy.Basic
            The expression that determines the state
        dependent : gotran.ODEObject
            If given the count of this expression will follow as a
            fractional count based on the count of the dependent object
        solve_flags : dict
            Flags that are passed directly to sympy.solve
        """

        from modelparameters.sympytools import symbols_from_expr

        state = self._expect_state(state)

        # Check the sympy flags
        if solve_flags.get("dict") or solve_flags.get("set"):
            error("Cannot use dict=True or set=True as sympy_flags")

        # Check that there are no expressions that are dependent on the state
        for sym in symbols_from_expr(expr, include_derivatives=True):
            if (state.sym is not sym) and (state.sym in sym.atoms()):
                error(
                    "{0}, a sub expression of the expression, cannot depend "
                    "on the state for which we try to solve for.".format(sym)
                )

        # Try solve the passed expr
        try:
            solved_expr = sp.solve(expr, state.sym)
        except:
            error("Could not solve the passed expression")

        assert isinstance(solved_expr, list)

        # FIXME: Add possibilities to choose solution?
        if len(solved_expr) != 1:
            error("expected only 1 solution")

        # Unpack the solution
        solved_expr = solved_expr[0]

        self.add_state_solution(state, solved_expr, dependent)

    def add_state_solution(self, state, expr, dependent=None):
        """
        Add a solution expression for a state

        Arguments
        ---------
        state : gotran.State, AppliedUndef
            The State that is solved
        expr : sympy.Basic
            The expression that determines the state
        dependent : gotran.ODEObject
            If given the count of this expression will follow as a
            fractional count based on the count of the dependent object
        """

        state = self._expect_state(state)

        if "d{0}_dt".format(state.name) in self.ode_objects:
            error(
                "Cannot registered a state solution for a state "
                "that has a state derivative registered."
            )

        if "alg_{0}_0".format(state.name) in self.ode_objects:
            error(
                "Cannot registered a state solution for a state "
                "that has an algebraic expression registered."
            )

        # Create a StateSolution in the present component
        obj = StateSolution(state, expr, dependent)

        self._register_component_object(obj)

    def add_intermediate(self, name, expr, dependent=None):
        """
        Register an intermediate math expression

        Arguments
        ---------
        name : str
            The name of the expression
        expr : sympy.Basic, scalar
            The expression
        dependent : gotran.ODEObject
            If given the count of this expression will follow as a
            fractional count based on the count of the dependent object
        """

        # Create an Intermediate in the present component
        timer = Timer("Add intermediate")
        expr = Intermediate(name, expr, dependent)

        self._register_component_object(expr, dependent)

        return expr.sym

    def add_derivative(self, der_expr, dep_var, expr, dependent=None):
        """
        Add a derivative expression

        Arguments
        ---------
        der_expr : gotran.Expression, gotran.State, sympy.AppliedUndef
            The Expression or State which is differentiated
        dep_var : gotran.State, gotran.Time, gotran.Expression, sympy.AppliedUndef, sympy.Symbol
            The dependent variable
        expr : sympy.Basic
            The expression which the differetiation should be equal
        dependent : gotran.ODEObject
            If given the count of this expression will follow as a
            fractional count based on the count of the dependent object
        """
        timer = Timer("Add derivatives")

        if isinstance(der_expr, AppliedUndef):
            name = sympycode(der_expr)
            der_expr = self.root.present_ode_objects.get(name)

            if der_expr is None:
                error("{0} is not registered in this ODE".format(name))

        if isinstance(dep_var, (AppliedUndef, sp.Symbol)):
            name = sympycode(dep_var)
            dep_var = self.root.present_ode_objects.get(name)

            if dep_var is None:
                error("{0} is not registered in this ODE".format(name))

        # Check if der_expr is a State
        if isinstance(der_expr, State):
            self._expect_state(der_expr)
            obj = StateDerivative(der_expr, expr, dependent)

        else:

            # Create a DerivativeExpression in the present component
            obj = DerivativeExpression(der_expr, dep_var, expr, dependent)

        self._register_component_object(obj, dependent)

        return obj.sym

    def add_algebraic(self, state, expr, dependent=None):
        """
        Add an algebraic expression which relates a State with an
        expression which should equal to 0

        Arguments
        ---------
        state : gotran.State
            The State which the algebraic expression should determine
        expr : sympy.Basic
            The expression that should equal 0
        dependent : gotran.ODEObject
            If given the count of this expression will follow as a
            fractional count based on the count of the dependent object
        """

        state = self._expect_state(state)

        if "d{0}_dt".format(state.name) in self.ode_objects:
            error(
                "Cannot registered an algebraic expression for a state "
                "that has a state derivative registered."
            )

        if state.is_solved:
            error(
                "Cannot registered an algebraic expression for a state "
                "which is registered solved."
            )

        # Create an AlgebraicExpression in the present component
        obj = AlgebraicExpression(state, expr, dependent)

        self._register_component_object(obj, dependent)

    @property
    def states(self):
        """
        Return a list of all states in the component and its children
        """
        return [state for state in iter_objects(self, False, False, False, State)]

    @property
    def full_states(self):
        """
        Return a list of all states in the component and its children that are
        not solved and determined by a state expression
        """
        return [
            expr.state for expr in self.state_expressions if not expr.state.is_solved
        ]

    @property
    def full_state_vector(self):
        """
        Return a sympy column vector with all full states
        """
        states = self.full_states

        return sp.Matrix(len(states), 1, lambda i, j: states[i].sym)

    @property
    def parameters(self):
        """
        Return a list of all parameters in the component
        """
        return ode_objects(self, Parameter)

    @property
    def intermediates(self):
        """
        Return a list of all intermediates
        """
        return [obj for obj in iter_objects(self, False, False, False, Intermediate)]

    @property
    def state_expressions(self):
        """
        Return a list of state expressions
        """
        return sorted(
            (obj for obj in iter_objects(self, False, False, False, StateExpression))
        )

    @property
    def rate_expressions(self):
        """
        Return a list of rate expressions
        """
        return [obj for obj in iter_objects(self, False, False, False, RateExpression)]

    @property
    def components(self):
        """
        Return a list of all child components in the component
        """
        return ode_components(self)

    @property
    def comments(self):
        return [com for comp in ode_components(self) for com in comp._local_comments]

    @property
    def root(self):
        """
        Return the root ODE component (the ode)
        """
        present = self
        while present != present.parent:
            present = present.parent

        return present

    @property
    def is_finalized(self):
        return self._is_finalized

    @property
    def num_states(self):
        """
        Return the number of all states
        """
        return len(self.states)

    @property
    def num_full_states(self):
        """
        Return the number of all full states
        """
        return len(self.full_states)

    @property
    def num_parameters(self):
        """
        Return the number of all parameters
        """
        return len(self.parameters)

    @property
    def num_intermediates(self):
        """
        Return the number of all intermediates
        """
        return len(self.intermediates)

    @property
    def num_state_expressions(self):
        """
        Return the number state expressions
        """
        return len(self.state_expressions)

    @property
    def num_rate_expressions(self):
        """
        Return the number rate expressions
        """
        return len(self.rate_expressions)

    @property
    def num_components(self):
        """
        Return the number of all components including it self
        """
        return len(self.components)

    @property
    def is_complete(self):
        """
        True if the component and all its children are locally complete
        """
        return self.is_locally_complete and all(
            child.is_complete for child in list(self.children.values())
        )

    @property
    def is_locally_complete(self):
        """
        True if the number of non-solved states are the same as the number
        of registered state expressions
        """
        num_local_states = sum(
            1
            for obj in self.ode_objects
            if isinstance(obj, State) and not obj.is_solved
        )

        return num_local_states == len(self._local_state_expressions)

    def __setattr__(self, name, value):
        """
        A magic function which will register expressions and simpler
        state expressions
        """

        from modelparameters.sympytools import symbols_from_expr

        # If we are registering a protected attribute or an attribute
        # during construction, just add it to the dict
        if name[0] == "_" or not self._allow_magic_attributes:
            self.__dict__[name] = value
            return

        # If no expression is registered
        if (not isinstance(value, scalars + (sp.Number,))) and not (
            isinstance(value, sp.Basic) and symbols_from_expr(value)
        ):
            debug(
                "Not registering: {0} as attribut. It does not contain "
                "any symbols or scalars.".format(name)
            )

            # FIXME: Should we raise an error?
            return

        # Check for special expressions
        expr, TYPE = special_expression(name, self.root)

        if TYPE == INTERMEDIATE:
            self.add_intermediate(name, value)

        elif TYPE == DERIVATIVE_EXPRESSION:

            # Try getting corresponding ODEObjects
            expr_name, var_name = expr.groups()
            expr_obj = self.root.present_ode_objects.get(expr_name)
            var_obj = self.root.present_ode_objects.get(var_name)

            # If the expr or variable is not declared in this ODE we
            # register an intermediate
            if expr_obj is None or var_obj is None:
                self.add_intermediate(name, value)

            else:
                self.add_derivative(expr_obj, var_obj, value)

        elif TYPE == STATE_SOLUTION_EXPRESSION:
            self.add_state_solution(expr, value)

        elif TYPE == ALGEBRAIC_EXPRESSION:

            # Try getting corresponding ODEObjects
            var_name, = expr.groups()
            var_obj = self.root.present_ode_objects.get(var_name)

            if var_obj is None:
                self.add_intermediate(name, value)
            else:
                self.add_algebraic(var_obj, value)

    def __call__(self, name):
        """
        Return a child component, if the component does not excist, create
        and add one
        """
        check_arg(name, str)

        # If the requested component is the root component
        if self == self.root and name == self.root.name:
            comp = self.root
        else:
            comp = self.children.get(name)

        if comp is None:
            comp = self.add_component(name)
            debug("Adding '{0}' component to {1}".format(name, self))
        else:
            self.root._present_component = comp.name

        return comp

    def _expect_state(self, state, allow_state_solution=False, only_local_states=False):
        """
        Help function to check an argument which should be expected
        to be a state
        """

        if allow_state_solution:
            allowed = (State, StateSolution)
        else:
            allowed = (State,)

        if isinstance(state, AppliedUndef):
            name = sympycode(state)
            state = self.root.present_ode_objects.get(name)

            if state is None:
                error("{} is not registered in this ODE".format(name))

            if only_local_states and not (
                state in self.states
                or (state in self.intermediates and allow_state_solution)
            ):
                error("{} is not registered in component {}".format(name, self.name))

        check_arg(state, allowed, 0)

        if isinstance(state, State) and state.is_solved:
            error(
                "Cannot registered a state expression for a state "
                "which is registered solved."
            )

        return state

    def _register_component_object(self, obj, dependent=None):
        """
        Register an ODEObject to the component
        """

        if self._is_finalized:
            error(
                "Cannot add {0} {1} to component {2} it is "
                "already finalized.".format(obj.__class__.__name__, obj, self)
            )

        self._check_reserved_wordings(obj)

        # Register the object in the root ODE,
        # (here all duplication checks and expression expansions are done)
        self.root.register_ode_object(obj, self, dependent)

        # If registering a StateExpression
        if isinstance(obj, StateExpression):

            if self.rates and not self._is_finalizing:
                error(
                    "A component cannot have both state expressions "
                    "(derivative and algebraic expressions) and rate "
                    "expressions."
                )

            if obj.state in self._local_state_expressions:
                error(
                    "A StateExpression for state {0} is already registered "
                    "in this component.".format(obj.state.name)
                )

            # Check that the state is registered in this component
            state_obj = self.ode_objects.get(obj.state.name)
            if not isinstance(state_obj, State):
                error(
                    "The state expression {0} defines state {1}, which is "
                    "not registered in the {2} component.".format(obj, obj.state, self)
                )

            self._local_state_expressions[obj.state] = obj

        # If obj is Intermediate register it as an attribute so it can be used
        # later on.
        if isinstance(obj, (State, Parameter, IndexedObject, Expression)):

            # If indexed expression or object register the basename as a dict
            if isinstance(obj, (IndexedExpression, IndexedObject)):
                if obj.basename in self.__dict__ and isinstance(
                    self.__dict__[obj.basename], dict
                ):
                    self.__dict__[obj.basename][obj.indices] = obj.sym
                else:
                    # FIXME: Initialize unused indices with zero
                    self.__dict__[obj.basename] = {obj.indices: obj.sym}

            # Register symbol, overwrite any already excisting symbol
            else:
                self.__dict__[obj.name] = obj.sym

        # Register the object
        self.ode_objects.append(obj)

    def _check_reserved_wordings(self, obj):
        if obj.name in _all_keywords:
            error(
                "Cannot register a {0} with a computer language "
                "keyword name: {1}".format(obj.__class__.__name__, obj.name)
            )

        # Check for reserved Expression wordings
        # if isinstance(obj, Expression):
        #    if re.search(_derivative_name_template, obj.name) \
        #           and not isinstance(obj, Derivatives):
        #        error("The pattern d{{name}}_dt is reserved for derivatives. "
        #              "However {0} is not a Derivative.".format(obj.name))
        #
        #    if re.search(_algebraic_name_template, obj.name) \
        #           and not isinstance(obj, AlgebraicExpression):
        #        error("The pattern {alg_{{name}}_0 is reserved for algebraic "\
        #              "expressions, however {1} is not an AlgebraicExpression."\
        #              .format(obj.name))

    def _add_rates(self, states, rate_matrix):
        """
        Use a rate matrix to set rates between states

        Arguments
        ---------
        states : list of States, tuple of two lists of States
            If one list is passed the rates should be a square matrix
            and the states list determines the order of the row and column of
            the matrix. If two lists are passed the first determines the states
            in the row and the second the states in the column of the Matrix
        rates_matrix : sympy.MatrixBase
            A sympy.Matrix of the rate expressions between the states given in
            the states argument
        """

        check_arg(states, (tuple, list), 0, ODEComponent._add_rates)
        check_arg(rate_matrix, sp.MatrixBase, 1, ODEComponent._add_rates)

        # If list
        if isinstance(states, list):
            states = (states, states)

        # else tuple
        elif len(states) != 2 and not all(
            isinstance(list_of_states, list) for list_of_states in states
        ):
            error("expected a tuple of 2 lists with states as the " "states argument")

        # Get all states associated with this Markov model
        local_states = self.states

        # Check index arguments
        # for list_of_states in states:
        #    print list_of_states, local_states
        #    if not all(state in local_states for state in list_of_states):
        #        error("Expected the states arguments to be States in "\
        #              "the Markov model")

        # Check that the length of the state lists corresponds with the shape of
        # the rate matrix
        if rate_matrix.shape[0] != len(states[0]) or rate_matrix.shape[1] != len(
            states[1]
        ):
            error("Shape of rates does not match given states")

        for i, state_i in enumerate(states[0]):
            for j, state_j in enumerate(states[1]):
                value = rate_matrix[i, j]

                # If 0 as rate
                if (isinstance(value, scalars) and value == 0) or (
                    isinstance(value, sp.Basic) and value.is_zero
                ):
                    continue

                if state_i == state_j:
                    error("Cannot have a nonzero rate value between the " "same states")

                # Assign the rate
                self._add_single_rate(state_i, state_j, value)

    def _add_single_rate(self, to_state, from_state, expr):
        """
        Add a single rate expression
        """

        if self.state_expressions:
            error(
                "A component cannot have both state expressions (derivative "
                "and algebraic expressions) and rate expressions."
            )

        check_arg(expr, scalars + (sp.Basic,), 2, ODEComponent._add_single_rate)

        expr = sp.sympify(expr)

        to_state = self._expect_state(
            to_state, allow_state_solution=True, only_local_states=True
        )
        from_state = self._expect_state(
            from_state, allow_state_solution=True, only_local_states=True
        )

        if to_state == from_state:
            error("The two states cannot be the same.")

        if (to_state.sym, from_state.sym) in self.rates:
            error(
                "Rate between state {0} and {1} is already "
                "registered.".format(from_state, to_state)
            )

        # FIXME: It should also not be possible to include other
        # states in the markov model, right?
        syms_expr = symbols_from_expr(expr)
        if to_state.sym in syms_expr or from_state.sym in syms_expr:
            error(
                "The rate expression cannot be dependent on the " "states it connects."
            )

        # Create a RateExpression
        obj = RateExpression(to_state, from_state, expr)

        self._register_component_object(obj)

        # Store the rate sym in the rate dict
        self.rates._register_single_rate(to_state.sym, from_state.sym, obj.sym)

    def finalize_component(self):
        """
        Called whenever the component should be finalized
        """
        if self._is_finalized:
            return

        # A flag to allow adding of for example Markov model rates
        self._is_finalizing = True

        # If component is a Markov model
        if self.rates:
            self._finalize_markov_model()

        if not self.is_locally_complete:
            incomplete_states = []
            for obj in self.ode_objects:
                if isinstance(obj, State):
                    if not obj in self._local_state_expressions:
                        incomplete_states.append(obj)

            incomplete_state_names = [s.name for s in incomplete_states]

            msg = "Cannot finalize component '{0}'. ".format(self)
            msg += "Missing time derivatives for the following states: {}".format(', '.join(incomplete_state_names))

            error(msg)

        self._is_finalizing = False
        self._is_finalized = True

    def _finalize_markov_model(self):
        """
        Finalize a markov model
        """

        # Derivatives
        states = self.states
        derivatives = defaultdict(lambda: sp.sympify(0.0))
        rate_check = defaultdict(lambda: 0)

        # Build rate information and check that each rate is added in a
        # symetric way
        used_states = [0] * self.num_states
        for rate in self.rate_expressions:

            # Get the states
            to_state, from_state = rate.states

            # Add to derivatives of the two states
            derivatives[from_state] -= rate.sym * from_state.sym
            derivatives[to_state] += rate.sym * from_state.sym

            if isinstance(from_state, StateSolution):
                from_state = from_state.state

            if isinstance(to_state, StateSolution):
                to_state = to_state.state

            # Register rate
            ind_from = states.index(from_state)
            ind_to = states.index(to_state)
            ind_tuple = (min(ind_from, ind_to), max(ind_from, ind_to))
            rate_check[ind_tuple] += 1

            used_states[ind_from] = 1
            used_states[ind_to] = 1

        # Check used states
        if 0 in used_states:
            error(
                "No rate registered for state {0}".format(states[used_states.find(0)])
            )

        # Check rate symetry
        for (ind_from, ind_to), times in list(rate_check.items()):
            if times != 2:
                error(
                    "Only one rate between the states {0} and {1} was "
                    "registered, expected two.".format(states[ind_from], states[ind_to])
                )

        # Add derivatives
        for state in states:

            # Skip solved states
            if not isinstance(state, State) or state.is_solved:
                continue

            self.add_derivative(state, state.time.sym, derivatives[state])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Johan Hake (hake.dev@gmail.com)"
__date__ = "2013-05-07 -- 2015-06-04"
__copyright__ = "Copyright (C) 2013 " + __author__
__license__ = "GNU LGPL Version 3.0 or later"

from gotran import *
from gotran.model.loadmodel import load_ode
from gotran.model.odeobjects import Comment
from collections import deque
from modelparameters.parameterdict import *


def join_indent_and_split(code, baseindent=4):

    code = deque(code)
    output = [" " * baseindent + code.popleft()] if code else [""]
    while code:
        if len(output[-1]) + 2 + len(code[0]) > 80:
            output.append(" " * baseindent + code.popleft())
        else:
            output[-1] += ", " + code.popleft()

    return "\n".join(output)


def gotranprobe(filename, params):

    set_log_level(WARNING)
    ode = load_ode(filename)
    set_log_level(INFO)

    if params.flat_view:
        print()
        print("  Components({}):".format(len(ode.components)))
        print(join_indent_and_split([comp.name for comp in ode.components]))
        print()
        print("  States({}):".format(len(ode.full_states)))
        print(join_indent_and_split([state.name for state in ode.full_states]))
        if len(ode.parameters) > 0:
            print()
            print("  Parameters({}):".format(len(ode.parameters)))
        print(join_indent_and_split([param.name for param in ode.parameters]))
        print()
        print("  Intermediates({}): ".format(len(ode.intermediates)))
        print(join_indent_and_split([interm.name for interm in ode.intermediates]))
    else:
        for comp in ode.components:
            if comp.states or comp.parameters or comp.intermediates:
                print("\n  {0}:".format(comp.name))
                if comp.full_states:
                    print(
                        "    States:        ",
                        ", ".join(state.name for state in comp.full_states),
                    )
                if comp.parameters:
                    print(
                        "    Parameters:    ",
                        ", ".join(param.name for param in comp.parameters),
                    )
                if comp.intermediates:
                    print(
                        "    Intermediates: ",
                        ", ".join(interm.name for interm in comp.intermediates),
                    )
                    print(
                        "    Dependencies:  ",
                        ", ".join(
                            set(
                                dep.name
                                for interm in comp.intermediates
                                for dep in ode.expression_dependencies[interm]
                                if not isinstance(dep, Comment)
                                and dep not in comp.full_states + comp.parameters
                            )
                        ),
                    )


def main():
    import sys, os

    params = ParameterDict(
        flat_view=Param(True, description="List all objects in a flat view")
    )
    params.parse_args(usage="usage: %prog FILE [options]")

    if len(sys.argv) < 2:
        raise RuntimeError("Expected a single gotran file argument")

    if not os.path.isfile(sys.argv[1]):
        raise IOError("Expected the argument to be a file")

    file_name = sys.argv[1]
    gotranprobe(file_name, params)


if __name__ == "__main__":
    main()



# Copyright (C) 2013 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

__all__ = ["CodeComponent"]

# System imports
from collections import OrderedDict, deque, defaultdict
from sympy.core.function import AppliedUndef
import sys
from functools import cmp_to_key

# ModelParameters imports
from modelparameters.sympytools import sp
from modelparameters.codegeneration import sympycode

# Local imports
from gotran.common import (
    error,
    info,
    debug,
    check_arg,
    check_kwarg,
    scalars,
    Timer,
    warning,
    tuplewrap,
    parameters,
)
from gotran.model.odeobjects import (
    State,
    Parameter,
    IndexedObject,
    StateIndexedObject,
    ParameterIndexedObject,
    Comment,
    cmp,
)
from gotran.model.expressions import *
from gotran.model.odecomponent import ODEComponent
from gotran.model.ode import ODE

# FIXME: Remove our own cse, or move to this module?
from gotran.codegeneration.sympy_cse import cse

# from sympy import cse


class CodeComponent(ODEComponent):
    """
    A wrapper class around an ODE. Its primary purpose is to help
    generate code.

    The class alows extraction and manipulation of the ODE expressions.
    """

    @staticmethod
    def default_parameters():
        """
        Return the default parameters for code generation
        """
        return parameters.generation.code.copy()

    def __init__(
        self,
        name,
        ode,
        function_name,
        description,
        params=None,
        use_default_arguments=True,
        additional_arguments=None,
        **results
    ):
        """
        Creates a CodeComponent

        Arguments
        ---------
        name : str
            The name of the component. This str serves as the unique
            identifier of the Component.
        ode : gotran.ODE
            The parent component which need to be a ODE
        function_name : str
            The name of the generated function
        description : str
            A short description of what the code component are computing
        params : dict
            Parameters determining how the code should be generated
        use_default_arguments : bool
            If true state, time and parameters are expected to be used in the
            code component
        additional_arguments : list
            A list of str for additional arguments included in the signature.
        results : kwargs
            A dict of result expressions. The result expressions will
            be used to extract the body expressions for the code component.
        """
        params = params or {}
        additional_arguments = additional_arguments or []
        check_arg(ode, ODE)
        check_arg(name, str)
        check_arg(description, str)
        check_arg(function_name, str)
        check_kwarg(params, "params", dict)
        check_kwarg(use_default_arguments, "use_default_arguments", bool)
        check_kwarg(additional_arguments, "additional_arguments", list)

        super(CodeComponent, self).__init__(name, ode)

        # Turn off magic attributes, see ODEComponent.__setattr__
        # method
        self._allow_magic_attributes = False
        for result_name, result_expressions in list(results.items()):
            check_kwarg(
                result_expressions, result_name, list, itemtypes=(Expression, Comment)
            )

        # Store parameters
        self._params = self.default_parameters()
        self._params.update(params)

        # Store function name and description
        self.function_name = function_name
        self.description = description

        # Shapes for any indexed expressions or objects
        self.shapes = OrderedDict()

        # A map between expressions and recreated IndexedExpressions
        self.indexed_map = OrderedDict()

        # Init parameter or state replace dict
        if use_default_arguments:
            self.param_state_replace_dict = self._init_param_state_replace_dict()
        else:
            self.param_state_replace_dict = {}

        self.results = list(results.keys())
        # Recreate body expressions based on the given result_expressions
        if results:
            results, body_expressions = self._body_from_results(**results)
            self.body_expressions = self._recreate_body(body_expressions, **results)
        else:
            self.body_expressions = []

        # Store for later usage
        self.use_default_arguments = use_default_arguments
        self.additional_arguments = additional_arguments

    def add_indexed_expression(
        self, basename, indices, expr, add_offset=False, dependent=None, enum=None
    ):
        """
        Add an indexed expression using a basename and the indices

        Arguments
        ---------
        basename : str
            The basename of the indexed expression
        indices : int, tuple of int
            The fixed indices identifying the expression
        expr : sympy.Basic, scalar
            The expression.
        add_offset : bool
            Add offset to indices
        dependent : gotran.ODEObject
            If given the count of this expression will follow as a
            fractional count based on the count of the dependent object
        """
        # Create an IndexedExpression in the present component
        timer = Timer("Add indexed expression")
        indices = tuplewrap(indices)

        # Check that provided indices fit with the registered shape
        if len(self.shapes[basename]) > len(indices):
            error(
                "Shape mismatch between indices {0} and registered "
                "shape for {1}{2}".format(indices, basename, self.shapes[basename])
            )

        for dim, (index, shape_ind) in enumerate(zip(indices, self.shapes[basename])):
            if index >= shape_ind:
                error(
                    "Indices must be smaller or equal to the shape. Mismatch "
                    "in dim {0}: {1}>={2}".format(dim + 1, index, shape_ind)
                )

        # Create the indexed expression
        if enum is None:
            expr = IndexedExpression(
                basename,
                indices,
                expr,
                self.shapes[basename],
                self._params.array,
                add_offset,
                dependent,
                enum=enum,
            )
        elif isinstance(enum, State):
            state = enum
            expr = StateIndexedExpression(
                basename,
                indices,
                expr,
                state,
                self.shapes[basename],
                self._params.array,
                add_offset,
                dependent,
            )
        elif isinstance(enum, Parameter):
            parameter = enum
            expr = ParameterIndexedExpression(
                basename,
                indices,
                expr,
                parameter,
                self.shapes[basename],
                self._params.array,
                add_offset,
                dependent,
            )
        else:
            error("enum must be State or Parameter. Was {0}".format(type(enum)))
        self._register_component_object(expr, dependent)

        return expr.sym

    def add_indexed_object(self, basename, indices, add_offset=False):
        """
        Add an indexed object using a basename and the indices

        Arguments
        ---------
        basename : str
            The basename of the indexed expression
        indices : int, tuple of int
            The fixed indices identifying the expression
        add_offset : bool
            Add offset to indices
        """
        timer = Timer("Add indexed object")

        indices = tuplewrap(indices)

        # Check that provided indices fit with the registered shape
        if len(self.shapes[basename]) > len(indices):
            error(
                "Shape mismatch between indices {0} and registered "
                "shape for {1}{2}".format(indices, basename, self.shapes[basename])
            )

        for dim, (index, shape_ind) in enumerate(zip(indices, self.shapes[basename])):
            if index >= shape_ind:
                error(
                    "Indices must be smaller or equal to the shape. Mismatch "
                    "in dim {0}: {1}>={2}".format(dim + 1, index, shape_ind)
                )

        # Create IndexedObject
        obj = IndexedObject(
            basename, indices, self.shapes[basename], self._params.array, add_offset
        )
        self._register_component_object(obj)

        # Return the sympy version of the object
        return obj.sym

    def indexed_objects(self, *basenames):
        """
        Return a list of all indexed objects with the given basename,
        if no base names give all indexed objects are returned
        """
        if not basenames:
            basenames = list(self.shapes.keys())
        return [
            obj
            for obj in self.ode_objects
            if isinstance(obj, IndexedObject) and obj.basename in basenames
        ]

    def _init_param_state_replace_dict(self):
        """
        Create a parameter state replace dict based on the values in the
        global parameters
        """

        param_state_replace_dict = {}

        array_params = self._params["array"]
        param_repr = self._params["parameters"]["representation"]
        param_name = self._params["parameters"]["array_name"]
        param_offset = self._params["parameters"]["add_offset"]
        field_param_name = self._params["parameters"]["field_array_name"]
        field_param_offset = self._params["parameters"]["add_field_offset"]
        field_parameters = self._params["parameters"]["field_parameters"]

        # If empty
        # FIXME: Get rid of this by introducing a ListParam type in modelparameters
        if len(field_parameters) == 1 and field_parameters[0] == "":
            field_parameters = []

        for param in field_parameters:
            if not isinstance(self.root.present_ode_objects[param], Parameter):
                error(
                    "Field parameter '{0}' is not a parameter in the "
                    "'{1}'".format(param, self.root)
                )

        state_repr = self._params["states"]["representation"]
        state_name = self._params["states"]["array_name"]
        state_offset = self._params["states"]["add_offset"]
        time_name = self._params["time"]["name"]
        dt_name = self._params["dt"]["name"]

        # Create a map between states, parameters and the corresponding
        # IndexedObjects
        param_state_map = OrderedDict(
            [("states", OrderedDict()), ("parameters", OrderedDict())]
        )

        # Add states
        states = param_state_map["states"]
        for ind, state in enumerate(self.root.full_states):
            states[state] = StateIndexedObject(
                state_name,
                ind,
                state,
                (self.root.num_full_states,),
                array_params,
                state_offset,
            )

        # Add parameters
        parameters = param_state_map["parameters"]
        for ind, param in enumerate(self.root.parameters):

            if param.name in field_parameters:
                basename = field_param_name
                index = field_parameters.index(param.name)
                shape = (len(field_parameters),)
                offset = field_param_offset
            else:
                basename = param_name
                index = ind
                shape = (self.num_parameters,)
                offset = param_offset

            parameters[param] = ParameterIndexedObject(
                basename, index, param, shape, array_params, offset
            )

        # If not having named parameters
        if param_repr == "numerals":
            param_state_replace_dict.update(
                (param.sym, param.init) for param in self.root.parameters
            )
        elif param_repr == "array":
            self.shapes[param_name] = (self.root.num_parameters,)
            if field_parameters:
                self.shapes["field_" + param_name] = (len(field_parameters),)
            param_state_replace_dict.update(
                (param.sym, indexed.sym)
                for param, indexed in list(param_state_map["parameters"].items())
            )

        if state_repr == "array":
            self.shapes[state_name] = (self.root.num_full_states,)
            param_state_replace_dict.update(
                (state.sym, indexed.sym)
                for state, indexed in list(param_state_map["states"].items())
            )

        param_state_replace_dict[self.root._time.sym] = sp.Symbol(time_name)
        param_state_replace_dict[self.root._dt.sym] = sp.Symbol(dt_name)

        self.indexed_map.update(param_state_map)

        # return  dicts
        return param_state_replace_dict

    def _body_from_results(self, **results):
        """
        Returns a sorted list of body expressions all used in the result expressions
        """

        # Store results
        self.results = list(results.keys())

        if not results:
            return {}, []

        # Check if we are using common sub expressions for body
        if self._params["body"]["use_cse"]:
            return self._body_from_cse(**results)
        else:
            return self._body_from_dependencies(**results)

    def _expanded_result_expressions(self, **results):

        # Extract all result expressions
        orig_result_expressions = sum(list(results.values()), [])

        # A map between result expression and result name
        result_names = dict(
            (result_expr, result_name)
            for result_name, result_exprs in list(results.items())
            for result_expr in result_exprs
        )

        # The expanded result expressions
        expanded_result_exprs = [
            self.root.expanded_expression(obj) for obj in orig_result_expressions
        ]

        # Set shape for result expressions
        for result_name, result_expressions in list(results.items()):
            if result_name not in self.shapes:
                self.shapes[result_name] = (len(result_expressions),)

        return orig_result_expressions, result_names, expanded_result_exprs

    def _only_result_expressions(self, **results):

        (
            orig_result_expressions,
            result_names,
            expanded_result_exprs,
        ) = self._expanded_result_expressions(**results)

        body_expressions = []
        new_results = defaultdict(list)

    def _body_from_cse(self, **results):

        timer = Timer("Compute common sub expressions for {0}".format(self.name))

        (
            orig_result_expressions,
            result_names,
            expanded_result_exprs,
        ) = self._expanded_result_expressions(**results)

        state_offset = self._params["states"]["add_offset"]

        # Collect results and body_expressions
        body_expressions = []
        new_results = defaultdict(list)

        might_take_time = len(orig_result_expressions) >= 40

        if might_take_time:
            info(
                "Computing common sub expressions for {0}. Might take "
                "some time...".format(self.name)
            )
            sys.stdout.flush()

        # Call sympy common sub expression reduction
        cse_exprs, cse_result_exprs = cse(
            expanded_result_exprs, symbols=sp.numbered_symbols("cse_"), optimizations=[]
        )

        # Map the cse_expr to an OrderedDict
        cse_exprs = OrderedDict(cse_expr for cse_expr in cse_exprs)

        # Extract the symbols into a set for fast comparison
        cse_syms = set((sym for sym in cse_exprs))

        # Create maps between cse_expr and result expressions trying
        # to optimized the code by weaving in the result expressions
        # in between the cse_expr

        # A map between result expr and name and indices so we can
        # construct IndexedExpressions
        result_expr_map = defaultdict(list)

        # A map between last cse_expr used in a particular result expr
        # so that we can put the result expression right after the
        # last cse_expr it uses.
        last_cse_expr_used_in_result_expr = defaultdict(list)

        # A map between replaced cse_sym and result_expressions
        cse_sym_to_result_expr = {}

        # Result expressions that does not contain any cse_sym
        result_expr_without_cse_syms = []

        # A map between cse_sym and its substitutes
        cse_subs = {}

        for ind, (orig_result_expr, result_expr) in enumerate(
            zip(orig_result_expressions, cse_result_exprs)
        ):

            # Collect information so that we can recreate the result
            # expression from
            result_expr_map[result_expr].append(
                (
                    result_names[orig_result_expr],
                    orig_result_expr.indices
                    if isinstance(orig_result_expr, IndexedExpression)
                    else ind,
                )
            )

            # If result_expr does not contain any cse_sym
            if not any(cse_sym in cse_syms for cse_sym in result_expr.atoms()):
                result_expr_without_cse_syms.append(result_expr)

            else:

                # Get last cse_sym used in result expression
                last_cse_sym = sorted(
                    (cse_sym for cse_sym in result_expr.atoms() if cse_sym in cse_syms),
                    key=cmp_to_key(lambda a, b: cmp(int(a.name[4:]), int(b.name[4:]))),
                )[-1]

                if result_expr not in last_cse_expr_used_in_result_expr[last_cse_sym]:
                    last_cse_expr_used_in_result_expr[last_cse_sym].append(result_expr)

        debug(
            "Found {0} result expressions without any cse_syms.".format(
                len(result_expr_without_cse_syms)
            )
        )

        # print ""
        # print "LAST cse_syms:", last_cse_expr_used_in_result_expr.keys()

        cse_cnt = 0
        atoms = [state.sym for state in self.root.full_states]
        atoms.extend(param.sym for param in self.root.parameters)

        # Collecte what states and parameters has been used
        used_states = set()
        used_parameters = set()

        self.add_comment(
            "Common sub expressions for the body and the " "result expressions"
        )
        body_expressions.append(self.ode_objects[-1])

        # Register the common sub expressions as Intermediates
        for cse_sym, expr in list(cse_exprs.items()):

            # print cse_sym, expr

            # If the expression is just one of the atoms of the ODE we
            # skip the cse expressions but add a subs for the atom We
            # also skip Relationals and Piecewise as the type checking
            # in Piecewise otherwise kicks in and destroys things for
            # us.
            if expr in atoms or isinstance(
                expr, (sp.Piecewise, sp.relational.Relational, sp.relational.Boolean)
            ):
                cse_subs[cse_sym] = expr.xreplace(cse_subs)
            else:
                # Add body expression as an intermediate expression
                sym = self.add_intermediate(
                    "cse_{0}".format(cse_cnt), expr.xreplace(cse_subs)
                )
                obj = self.ode_objects.get(sympycode(sym))
                for dep in self.root.expression_dependencies[obj]:
                    if isinstance(dep, State):
                        used_states.add(dep)
                    elif isinstance(dep, Parameter):
                        used_parameters.add(dep)
                cse_subs[cse_sym] = sym
                cse_cnt += 1
                body_expressions.append(obj)

            # Check if we should add a result expressions
            if last_cse_expr_used_in_result_expr[cse_sym]:

                # Iterate over all registered result expr for this cse_sym
                for result_expr in last_cse_expr_used_in_result_expr.pop(cse_sym):

                    for result_name, indices in result_expr_map[result_expr]:

                        # Replace pure state and param expressions
                        # print cse_subs, result_expr
                        exp_expr = result_expr.xreplace(cse_subs)

                        sym = self.add_indexed_expression(
                            result_name, indices, exp_expr, add_offset=state_offset
                        )

                        expr = self.ode_objects.get(sympycode(sym))

                        for dep in self.root.expression_dependencies[expr]:
                            if isinstance(dep, State):
                                used_states.add(dep)
                            elif isinstance(dep, Parameter):
                                used_parameters.add(dep)

                        # Register the new result expression
                        new_results[result_name].append(expr)
                        body_expressions.append(expr)

        if might_take_time:
            info(" done")

        # Sort used state, parameters and expr
        self.used_states = sorted(used_states)
        self.used_parameters = sorted(used_parameters)

        return new_results, body_expressions

    def _body_from_dependencies(self, **results):

        timer = Timer("Compute dependencies for {0}".format(self.name))

        # Extract all result expressions
        result_expressions = sum(list(results.values()), [])

        # Check passed expressions
        ode_expr_deps = self.root.expression_dependencies
        exprs = set(result_expressions)
        not_checked = set()
        used_states = set()
        used_parameters = set()

        for expr in result_expressions:
            check_arg(
                expr, (Expression, Comment), context=CodeComponent._body_from_results
            )

            if isinstance(expr, Comment):
                continue

            # Collect dependencies
            for obj in ode_expr_deps[expr]:
                if isinstance(obj, (Expression, Comment)):
                    not_checked.add(obj)
                elif isinstance(obj, State):
                    used_states.add(obj)
                elif isinstance(obj, Parameter):
                    used_parameters.add(obj)

        # use list to make the order consistent
        not_checked_list = sorted(not_checked)

        # Collect all dependencies
        while len(not_checked_list) > 0:
            dep_expr = not_checked_list.pop()
            exprs.add(dep_expr)
            for obj in ode_expr_deps[dep_expr]:
                if isinstance(obj, (Expression, Comment)):
                    if obj not in exprs:
                        if not obj in not_checked_list:
                            not_checked_list.append(obj)
                elif isinstance(obj, State):
                    used_states.add(obj)
                elif isinstance(obj, Parameter):
                    used_parameters.add(obj)

        # Sort used state, parameters and expr
        self.used_states = sorted(used_states)
        self.used_parameters = sorted(used_parameters)

        # Return a sorted list of all collected expressions
        return results, sorted(exprs)

    def _recreate_body(self, body_expressions, **results):
        """
        Create body expressions based on the given result_expressions

        In this method are all expressions replaced with something that should
        be used to generate code. The parameters in:

            parameters["generation"]["code"]

        decides how parameters, states, body expressions and indexed expressions
        are represented.

        """

        if not (results or body_expressions):
            return

        for result_name, result_expressions in list(results.items()):
            check_kwarg(
                result_expressions,
                result_name,
                list,
                context=CodeComponent._recreate_body,
                itemtypes=(Expression, Comment),
            )

        # Extract all result expressions
        result_expressions = sum(list(results.values()), [])

        # A map between result expression and result name
        result_names = dict(
            (result_expr, result_name)
            for result_name, result_exprs in list(results.items())
            for result_expr in result_exprs
        )

        timer = Timer("Recreate body expressions for {0}".format(self.name))

        # Initialize the replace_dictionaries
        replace_dict = self.param_state_replace_dict
        der_replace_dict = {}

        # Get a copy of the map of where objects are used in and their
        # present dependencies so any updates done in these dictionaries does not
        # affect the original dicts
        object_used_in = defaultdict(set)
        for expr, used in list(self.root.object_used_in.items()):
            object_used_in[expr].update(used)

        expression_dependencies = defaultdict(set)
        for expr, deps in list(self.root.expression_dependencies.items()):
            expression_dependencies[expr].update(deps)

        # Get body parameters
        body_repr = self._params["body"]["representation"]
        optimize_exprs = self._params["body"]["optimize_exprs"]

        # Set body related variables if the body should be represented by an array
        if "array" in body_repr:
            body_name = self._params["body"]["array_name"]
            available_indices = deque()
            body_indices = []
            max_index = -1
            body_ind = 0
            index_available_at = defaultdict(list)
            if body_name == result_name:
                error("body and result cannot have the same name.")

            # Initiate shapes with inf
            self.shapes[body_name] = (float("inf"),)

        # Iterate over body expressions and recreate the different expressions
        # according to state, parameters, body and result expressions
        replaced_expr_map = OrderedDict()
        new_body_expressions = []

        present_ode_objects = dict(
            (state.name, state) for state in self.root.full_states
        )
        present_ode_objects.update(
            (param.name, param) for param in self.root.parameters
        )
        old_present_ode_objects = present_ode_objects.copy()

        def store_expressions(expr, new_expr):
            "Help function to store new expressions"

            timer = Timer(
                "Store expression while recreating body of {}".format(self.name)
            )

            # Update sym replace dict
            if isinstance(expr, Derivatives):
                der_replace_dict[expr.sym] = new_expr.sym
            else:
                replace_dict[expr.sym] = new_expr.sym

            # Store the new expression for later references
            present_ode_objects[expr.name] = new_expr
            replaced_expr_map[expr] = new_expr

            # Append the new expression
            new_body_expressions.append(new_expr)

            # Update dependency information
            if expr in object_used_in:
                for dep in object_used_in[expr]:
                    if dep in expression_dependencies:
                        expression_dependencies[dep].remove(expr)
                        expression_dependencies[dep].add(new_expr)

                object_used_in[new_expr] = object_used_in.pop(expr)

            if expr in expression_dependencies:
                expression_dependencies[new_expr] = expression_dependencies.pop(expr)

        self.add_comment("Recreated body expressions")

        # The main iteration over all body_expressions
        for expr in body_expressions:

            # 1) Comments
            if isinstance(expr, Comment):
                new_body_expressions.append(expr)
                continue

            assert isinstance(expr, Expression)

            # 2) Check for expression optimizations
            if not (optimize_exprs == "none" or expr in result_expressions):

                timer_opt = Timer(
                    "Handle expression optimization for {0}".format(self.name)
                )

                # If expr is just a number we exchange the expression with the
                # number
                if "numerals" in optimize_exprs and isinstance(expr.expr, sp.Number):
                    replace_dict[expr.sym] = expr.expr

                    # Remove information about this expr beeing used
                    for dep in object_used_in[expr]:
                        expression_dependencies[dep].remove(expr)
                    object_used_in.pop(expr)
                    continue

                # If the expr is just a symbol (symbol multiplied with a scalar)
                # we exchange the expression with the sympy expressions
                elif "symbols" in optimize_exprs and (
                    isinstance(expr.expr, (sp.Symbol, AppliedUndef))
                    or isinstance(expr.expr, sp.Mul)
                    and len(expr.expr.args) == 2
                    and isinstance(expr.expr.args[1], (sp.Symbol, AppliedUndef))
                    and expr.expr.args[0].is_number
                ):

                    # Add a replace rule based on the stored sympy expression
                    sympy_expr = expr.expr.xreplace(der_replace_dict).xreplace(
                        replace_dict
                    )

                    if isinstance(expr.sym, sp.Derivative):
                        der_replace_dict[expr.sym] = sympy_expr
                    else:
                        replace_dict[expr.sym] = sympy_expr

                    # Get exchanged repr
                    if isinstance(expr.expr, (sp.Symbol, AppliedUndef)):
                        name = sympycode(expr.expr)
                    else:
                        name = sympycode(expr.expr.args[1])

                    dep_expr = present_ode_objects[name]

                    # If using reused body expressions we need to update the
                    # index information so that the index previously available
                    # for this expressions gets available at the last expressions
                    # the present expression is used in.
                    if (
                        isinstance(dep_expr, IndexedExpression)
                        and dep_expr.basename == body_name
                        and "reused" in body_repr
                    ):
                        ind = dep_expr.indices[0]

                        # Remove available index information
                        dep_used_in = sorted(object_used_in[dep_expr])
                        for used_expr in dep_used_in:
                            if ind in index_available_at[used_expr]:
                                index_available_at[used_expr].remove(ind)

                        # Update with new indices
                        all_used_in = object_used_in[expr].copy()
                        all_used_in.update(dep_used_in)

                        for used_expr in sorted(all_used_in, reverse=True):
                            if used_expr in body_expressions:
                                index_available_at[used_expr].append(ind)
                                break

                    # Update information about this expr beeing used
                    for dep in object_used_in[expr]:
                        expression_dependencies[dep].remove(expr)
                        expression_dependencies[dep].add(dep_expr)

                    object_used_in.pop(expr)
                    continue

                del timer_opt

            # 3) General operations for all Expressions that are kept

            # Before we process the expression we check if any indices gets
            # available with the expr (Only applies for the "reused" option for
            # body_repr.)
            if "reused" in body_repr:

                # Check if any indices are available at this expression ind
                available_indices.extend(index_available_at[expr])

            # Store a map of old name this will preserve the ordering of
            # expressions with the same name, similar to how this is treated in
            # the actual ODE.
            present_ode_objects[expr.name] = expr
            old_present_ode_objects[expr.name] = expr

            # 4) Handle result expression
            if expr in result_expressions:

                timer_result = Timer(
                    "Handle result expressions for {0}".format(self.name)
                )

                # Get the result name
                result_name = result_names[expr]

                # If the expression is an IndexedExpression with the same basename
                # as the result name we just recreate it
                if (
                    isinstance(expr, IndexedExpression)
                    or isinstance(expr, StateIndexedExpression)
                    or isinstance(expr, ParameterIndexedExpression)
                ) and result_name == expr.basename:

                    new_expr = recreate_expression(expr, der_replace_dict, replace_dict)

                # Not an indexed expression
                else:

                    # Get index based on the original ordering
                    index = (result_expressions.index(expr),)
                    # Create the IndexedExpression
                    # NOTE: First replace any derivative expression replaces, then state and
                    # NOTE: params
                    if isinstance(expr, StateDerivative):
                        new_expr = StateIndexedExpression(
                            result_name,
                            index,
                            expr.expr.xreplace(der_replace_dict).xreplace(replace_dict),
                            expr.state,
                            (len(results[result_name]),),
                            array_params=self._params.array,
                        )
                    else:
                        new_expr = IndexedExpression(
                            result_name,
                            index,
                            expr.expr.xreplace(der_replace_dict).xreplace(replace_dict),
                            (len(results[result_name]),),
                            array_params=self._params.array,
                        )

                    if new_expr.basename not in self.indexed_map:
                        self.indexed_map[new_expr.basename] = OrderedDict()
                    self.indexed_map[new_expr.basename][expr] = new_expr

                    # Copy counter from old expression so it sort properly
                    new_expr._recount(expr._count)

                # Store the expressions
                store_expressions(expr, new_expr)

                del timer_result

            # 4) Handle indexed expression
            # All indexed expressions are just kept but recreated with updated
            # sympy expressions
            elif isinstance(expr, IndexedExpression):

                timer_indexed = Timer(
                    "Handle indexed expressions for {0}".format(self.name)
                )

                new_expr = recreate_expression(expr, der_replace_dict, replace_dict)

                # Store the expressions
                store_expressions(expr, new_expr)

                del timer_indexed

            # 5) If replacing all body exressions with an indexed expression
            elif "array" in body_repr:

                timer_body = Timer("Handle body expressions for {0}".format(self.name))

                # 5a) If we reuse array indices
                if "reused" in body_repr:

                    if available_indices:
                        ind = available_indices.popleft()
                    else:
                        max_index += 1
                        ind = max_index

                    # Check when present ind gets available again
                    for used_expr in sorted(object_used_in[expr], reverse=True):
                        if used_expr in body_expressions:
                            index_available_at[used_expr].append(ind)
                            break
                    else:
                        warning("SHOULD NOT COME HERE!")

                # 5b) No reuse of array indices. Here each index corresponds to
                #     a distinct body expression
                else:

                    ind = body_ind

                    # Increase body_ind
                    body_ind += 1

                # Create the IndexedExpression
                new_expr = IndexedExpression(
                    body_name,
                    ind,
                    expr.expr.xreplace(der_replace_dict).xreplace(replace_dict),
                    array_params=self._params.array,
                    enum=expr.name,
                )

                if body_name not in self.indexed_map:
                    self.indexed_map[body_name] = OrderedDict()
                self.indexed_map[body_name][expr] = new_expr

                # Copy counter from old expression so they sort properly
                new_expr._recount(expr._count)

                # Store the expressions
                store_expressions(expr, new_expr)

                del timer_body

            # 6) If the expression is just an ordinary body expression and we
            #    are using named representation of body
            else:

                timer_expr = Timer("Handle expressions for {0}".format(self.name))
                # If the expression is a state derivative we need to add a
                # replacement for the Derivative symbol
                if isinstance(expr, StateDerivative):
                    new_expr = Intermediate(
                        sympycode(expr.sym),
                        expr.expr.xreplace(der_replace_dict).xreplace(replace_dict),
                    )
                    new_expr._recount(expr._count)
                else:
                    new_expr = recreate_expression(expr, der_replace_dict, replace_dict)

                del timer_expr

                # Store the expressions
                store_expressions(expr, new_expr)

        # Store indices for any added arrays
        if "reused_array" == body_repr:

            if max_index > -1:
                self.shapes[body_name] = (max_index + 1,)
            else:
                self.shapes.pop(body_name)

        elif "array" == body_repr:

            if body_ind > 0:
                self.shapes[body_name] = (body_ind,)
            else:
                self.shapes.pop(body_name)

        # Store the shape of the added result expressions
        for result_name, result_expressions in list(results.items()):
            if result_name not in self.shapes:
                self.shapes[result_name] = (len(result_expressions),)

        return new_body_expressions



# Copyright (C) 2012 Johan Hake
#
# This file is part of Gotran.
#
# Gotran is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gotran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Gotran. If not, see <http://www.gnu.org/licenses/>.

# FIXME: This is a deprecated module. Most of it should be
# incorporated into other modules by now.

# System imports
from collections import deque, OrderedDict
import hashlib
import re
import sys
from distutils.version import LooseVersion as _V

# Model parametrs imports
from modelparameters.parameterdict import *
from modelparameters.sympytools import sp, iter_symbol_params_from_expr

_current_sympy_version = _V(sp.__version__)

# Local gotran imports
from gotran.model.ode import ODE
from gotran.model.odecomponent import ODEComponent, Comment
from gotran.common import check_arg, check_kwarg, info

from .sympy_cse import cse

_jacobian_pattern = re.compile("_([0-9]+)")


def _iszero(x):
    """Returns True if x is zero."""
    x = sp.sympify(x)
    return x.is_zero


def _default_params(exclude=None):
    exclude = exclude or []
    check_arg(exclude, list, itemtypes=str)

    # Add not implemented parameters to excludes
    exclude += ["max_terms", "parameter_contraction", "use_variables"]

    # Build a dict with allowed parameters
    params = {}

    if "use_state_names" not in exclude:
        # Use state names in code (compared to array with indices)
        params["use_state_names"] = Param(
            True,
            description="Use state names in code " "(compared to array with indices)",
        )

    if "use_parameter_names" not in exclude:
        # Use parameter names in code (compared to array with indices)
        params["use_parameter_names"] = Param(
            True,
            description="Use parameter names "
            "in code (compared to array with indices)",
        )

    if "keep_intermediates" not in exclude:
        # Keep all intermediates
        params["keep_intermediates"] = Param(
            True, description="Keep intermediates in code"
        )

    if "use_variables" not in exclude:
        # If True, code for altering variables are created
        # FIXME: Not used
        params["use_variables"] = Param(
            False, description="If True, code for altering variables are created"
        )

    if "parameter_contraction" not in exclude:
        # Find sub expressions of only parameters and create a dummy parameter
        # FIXME: Not used
        params["parameter_contraction"] = Param(
            False,
            description="Find sub expressions of only parameters "
            "and create a dummy parameter",
        )

    if "parameter_numerals" not in exclude:
        # Exchange all parameters with their initial numerical values
        params["parameter_numerals"] = Param(
            False,
            description="Exchange all parameters with their initial"
            " numerical values",
        )

    if "max_terms" not in exclude:
        # Split terms with more than max_terms into several evaluations
        # FIXME: Not used
        params["max_terms"] = ScalarParam(
            5,
            ge=2,
            description="Split terms with more than max_terms "
            "into several evaluations",
        )

    if "use_cse" not in exclude:
        # Use sympy common sub expression simplifications,
        # only when keep_intermediates is false
        params["use_cse"] = Param(
            False,
            description="Use sympy common sub expression "
            "simplifications, only when keep_intermediates is false",
        )

    if "generate_jacobian" not in exclude:
        # Generate code for the computation of the jacobian
        params["generate_jacobian"] = Param(
            True, description="Generate code for the computation of the jacobian"
        )

    if "transposed_jacobian" not in exclude:
        # Generate code for the computation of the jacobian
        params["transposed_jacobian"] = Param(
            False, description="The Jacobian is transposed"
        )

    if "generate_lu_factorization" not in exclude:
        # Generate code for the factorization of the jacobian
        params["generate_lu_factorization"] = Param(
            True, description="Generate code for the factorization of the jacobian"
        )

    if "generate_forward_backward_subst" not in exclude:
        # Generate code for the forward backward substitution code
        params["generate_forward_backward_subst"] = Param(
            True, description="Generate code for the forward backward substitution code"
        )

    if "generate_linearized_evaluation" not in exclude:
        # Generate code for linearized evaluation
        # For sympy versions lower than 0.7.2 linearized computation does not work
        default_linearized_evaluation = _current_sympy_version > _V("0.7.2")

        params["generate_linearized_evaluation"] = Param(
            default_linearized_evaluation,
            description="Generate code for linearized evaluation",
        )

    # Return the ParameterDict
    return ParameterDict(**params)


class ODERepresentation(object):
    """
    Intermediate ODE representation where various optimizations
    can be performed.
    """

    def __init__(self, ode, name="", **optimization):
        """
        Create an ODERepresentation

        Arguments
        ---------
        ode : ODE
            The ode to be represented
        name : str (optional)
            An argument which determines the name of the ode representation,
            making it possible to create different representations identified
            with different names.
        """
        check_arg(ode, ODE, 0)
        check_kwarg(name, "name", str)

        self.ode = ode

        # Store the name
        self._name = name if name else ode.name

        self.optimization = _default_params()

        self.optimization.update(optimization)
        self._symbol_subs = None
        self.index = lambda i: "[{0}]".format(i)

        # Init prefix info
        self._state_prefix = ""
        self._parameter_prefix = ""

        self._used_in_monitoring = None
        self._cse_subs = None
        self._jacobian_expr = None
        self._cse_subs_single_dy = None
        self._linear_terms = None
        self._jacobian_action_vec = None
        self._cse_jacobian_expr = None
        self._cse_jacobian_action_expr = None

    @property
    def used_in_monitoring(self):
        if self._used_in_monitoring is None:
            self._compute_monitor_cse()
        return self._used_in_monitoring

    @property
    def used_in_linear_dy(self):
        if self._linear_terms is None:
            self._compute_linearized_dy()

        return self._used_in_linear_dy

    @property
    def used_in_single_dy(self):
        if self._linear_terms is None:
            self._compute_linearized_dy()
        return self._used_in_single_dy

    def signature(self):
        # Create a unique signature
        h = hashlib.sha1()
        h.update(self.ode.signature() + repr(self.optimization))
        return h.hexdigest()

    def _compute_dy_cse(self):
        if self._cse_subs is not None:
            return

        ode = self.ode

        info("Calculating common sub expressions. May take some time...")
        sys.stdout.flush()
        # If we use cse we extract the sub expressions here and cache
        # information
        self._cse_subs, self._cse_derivative_expr = cse(
            [self.subs(expr) for der, expr in ode.get_derivative_expr(True)],
            symbols=sp.numbered_symbols("cse_"),
            optimizations=[],
        )

        cse_counts = [[] for i in range(len(self._cse_subs))]
        for i in range(len(self._cse_subs)):
            for j in range(i + 1, len(self._cse_subs)):
                if self._cse_subs[i][0] in self._cse_subs[j][1].atoms():
                    cse_counts[i].append(j)

            for j in range(len(self._cse_derivative_expr)):
                if self._cse_subs[i][0] in self._cse_derivative_expr[j].atoms():
                    cse_counts[i].append(j + len(self._cse_subs))

        # Store usage count
        # FIXME: Use this for more sorting!
        self._cse_counts = cse_counts

        info(" done")

    def _compute_monitor_cse(self):
        if self._used_in_monitoring is not None:
            return

        # Generate info about used states and parameters
        self._used_in_monitoring = dict(states=set(), parameters=set())

        ode = self.ode

        for name, obj in list(ode.monitored_intermediates.items()):
            for sym in iter_symbol_params_from_expr(obj.expanded_expr):

                if ode.has_state(sym):
                    self._used_in_monitoring["states"].add(sym.name)
                elif ode.has_parameter(sym):
                    self._used_in_monitoring["parameters"].add(sym.name)
                else:
                    # Skip if Symbols is not state or parameter
                    pass

        self._cse_monitored_subs, self._cse_monitored_expr = cse(
            [
                self.subs(obj.expanded_expr)
                for obj in list(ode.monitored_intermediates.values())
            ],
            symbols=sp.numbered_symbols("cse_monitored_"),
            optimizations=[],
        )

        cse_counts = [[] for i in range(len(self._cse_monitored_subs))]
        for i in range(len(self._cse_monitored_subs)):
            for j in range(i + 1, len(self._cse_monitored_subs)):
                if (
                    self._cse_monitored_subs[i][0]
                    in self._cse_monitored_subs[j][1].atoms()
                ):
                    cse_counts[i].append(j)

            for j in range(len(self._cse_monitored_expr)):
                if (
                    self._cse_monitored_subs[i][0]
                    in self._cse_monitored_expr[j].atoms()
                ):
                    cse_counts[i].append(j + len(self._cse_monitored_subs))

        # Store usage count
        # FIXME: Use this for more sorting!
        self._cse_monitored_counts = cse_counts

        self._used_in_monitoring["parameters"] = list(
            self._used_in_monitoring["parameters"]
        )

        self._used_in_monitoring["states"] = list(self._used_in_monitoring["states"])

    def _compute_jacobian(self):
        if self._jacobian_expr is not None:
            return

        ode = self.ode

        if ode.num_states > 10:
            info(
                "Calculating jacobian for {0} states. "
                "May take some time...".format(ode.num_states)
            )
            sys.stdout.flush()

        sym_map = OrderedDict()
        jacobi_expr = OrderedDict()
        for i, (ders, expr) in enumerate(ode.get_derivative_expr(True)):
            for j, state in enumerate(ode.states):
                F_ij = expr.diff(state.sym)

                # Only collect non zero contributions
                if F_ij:
                    jacobi_sym = sp.Symbol("j_{0}_{1}".format(i, j))
                    sym_map[i, j] = jacobi_sym
                    jacobi_expr[jacobi_sym] = F_ij
                    # print "[%d,%d] (%d) # [%s, %s]  \n%s" \
                    # % (i,j, len(F_ij.args), states[i], states[j], F_ij)

        # Create the Jacobian
        self._jacobian_mat = sp.SparseMatrix(
            ode.num_states, ode.num_states, lambda i, j: sym_map.get((i, j), 0)
        )

        if self.optimization.transposed_jacobian:
            self._jacobian_mat = self._jacobian_mat.transpose

        self._jacobian_expr = jacobi_expr

        if ode.num_states > 10:
            info(" done")

    def _compute_jacobian_action(self):
        self._compute_jacobian()
        if self._jacobian_action_vec is not None:
            return

        ode = self.ode

        # Create a state vector
        ode_states = sp.Matrix(ode.num_states, 1, lambda i, j: self.ode.states[i])
        self._jacobian_action_vec = self._jacobian_mat * ode_states
        self._jacobian_action_expr = [
            jac_ac_expr.subs(self._jacobian_expr)
            for jac_ac_expr in self._jacobian_action_vec
        ]

    def _compute_jacobian_action_cse(self):
        self._compute_jacobian_action()

        if not self.optimization.use_cse or self._cse_jacobian_action_expr is not None:
            return

        info(
            "Calculating jacobian action common sub expressions for {0} entries. "
            "May take some time...".format(len(self._jacobian_action_expr))
        )
        sys.stdout.flush()

        # If we use cse we extract the sub expressions here and cache
        # information
        self._cse_jacobian_action_subs, self._cse_jacobian_action_expr = cse(
            [self.subs(expr) for expr in self._jacobian_action_expr],
            symbols=sp.numbered_symbols("cse_jacobian_action_"),
            optimizations=[],
        )

        cse_jacobian_action_counts = [
            [] for i in range(len(self._cse_jacobian_action_subs))
        ]

        for i in range(len(self._cse_jacobian_action_subs)):
            for j in range(i + 1, len(self._cse_jacobian_action_subs)):
                if (
                    self._cse_jacobian_action_subs[i][0]
                    in self._cse_jacobian_action_subs[j][1].atoms()
                ):
                    cse_jacobian_action_counts[i].append(j)

            for j in range(len(self._cse_jacobian_action_expr)):
                if (
                    self._cse_jacobian_action_subs[i][0]
                    in self._cse_jacobian_action_expr[j].atoms()
                ):
                    cse_jacobian_action_counts[i].append(
                        j + len(self._cse_jacobian_action_subs)
                    )

        # Store usage count
        # FIXME: Use this for more sorting!
        self._cse_jacobian_action_counts = cse_jacobian_action_counts

        info(" done")

    def _compute_jacobian_cse(self):

        self._compute_jacobian()

        if not self.optimization.use_cse or self._cse_jacobian_expr is not None:
            return

        info(
            "Calculating jacobi common sub expressions for {0} entries. "
            "May take some time...".format(len(self._jacobian_expr))
        )
        sys.stdout.flush()

        # If we use cse we extract the sub expressions here and cache
        # information
        self._cse_jacobian_subs, self._cse_jacobian_expr = cse(
            [self.subs(expr) for expr in list(self._jacobian_expr.values())],
            symbols=sp.numbered_symbols("cse_jacobian_"),
            optimizations=[],
        )

        cse_jacobian_counts = [[] for i in range(len(self._cse_jacobian_subs))]
        for i in range(len(self._cse_jacobian_subs)):
            for j in range(i + 1, len(self._cse_jacobian_subs)):
                if (
                    self._cse_jacobian_subs[i][0]
                    in self._cse_jacobian_subs[j][1].atoms()
                ):
                    cse_jacobian_counts[i].append(j)

            for j in range(len(self._cse_jacobian_expr)):
                if self._cse_jacobian_subs[i][0] in self._cse_jacobian_expr[j].atoms():
                    cse_jacobian_counts[i].append(j + len(self._cse_jacobian_subs))

        # Store usage count
        # FIXME: Use this for more sorting!
        self._cse_jacobian_counts = cse_jacobian_counts

        info(" done")

    def _compute_symbolic_factorization_of_jacobian(self):

        if self._jacobian_expr is None:
            self._compute_jacobian_cse()

        # Get jacobian
        AA = self._jacobian_mat

        if not AA.is_square:
            raise NonSquareMatrixError()

        # Get copy
        n = AA.rows
        A = AA[:, :]
        p = []

        n2 = n * n

        nnz = 0
        for i in range(n):
            # Add dummy value at the diagonal if zero
            if _iszero(A[i, i]):
                A[i, i] = 1
            for j in range(n):
                nnz += not _iszero(A[i, j])

        print("Num non zeros in jacobian:", nnz, nnz * 1.0 / n2 * 100, "%")

        # A map between old symbols and new. The values in this dict corresponds to
        # where an old symbol is used. If the length of the value is 1 it is only
        # used once and once the value is used it can be freed.
        old_new = OrderedDict()
        new_old = OrderedDict()
        operations = []
        new_count = 0

        global zero_operations
        zero_operations = 0

        def update_entry(new_count, i, j, k):
            global zero_operations
            if _iszero(A[i, k] * A[k, j]):
                zero_operations += 1
                return new_count

            # Store operation
            if i != 0:
                operations.append(
                    "jac[{i}*{n}+{j}] -= jac[{i}*{n}+{k}]*jac[{k}*{n}+{j}]".format(
                        i=i, j=j, k=k, n=n
                    )
                )
            else:
                operations.append(
                    "jac[{j}] -= jac[{k}]*jac[{k}*{n}+{j}]".format(j=j, k=k, n=n)
                )

            new_sym = sp.Symbol("j_{0}_{1}:{2}".format(i, j, new_count))
            new_old[new_sym] = A[i, j] - A[i, k] * A[k, j]
            for old_sym in [A[i, j], A[i, k], A[k, j]]:
                storage = old_new.get(old_sym)
                if storage is None:
                    storage = set()
                    old_new[old_sym] = storage
                storage.add(new_sym)

            # Change entry to the new symbol
            A[i, j] = new_sym
            new_count += 1

            return new_count

        # factorization
        for j in range(n):
            for i in range(j):
                for k in range(i):
                    new_count = update_entry(new_count, i, j, k)
            pivot = -1
            for i in range(j, n):
                for k in range(j):
                    new_count = update_entry(new_count, i, j, k)

                # find the first non-zero pivot, includes any expression
                if pivot == -1 and not _iszero(A[i, j]):
                    pivot = i
            if pivot < 0:
                # this result is based on iszerofunc's analysis of the
                # possible pivots, so even though the element may not be
                # strictly zero, the supplied iszerofunc's evaluation gave
                # True
                raise ValueError("No nonzero pivot found; inversion failed.")

            if pivot != j:  # row must be swapped
                A.row_swap(pivot, j)
                p.append([pivot, j])
                print("Pivoting!!")

            scale = 1 / A[j, j]
            for i in range(j + 1, n):
                if _iszero(A[i, j]):
                    zero_operations += 1
                    continue

                # Store operation
                operations.append(
                    "jac[{i}*{n}+{j}] /= jac[{j}*{n}+{j}]".format(i=i, n=n, j=j)
                )

                # Create new symbol and store the representation
                new_sym = sp.Symbol("j_{0}_{1}:{2}".format(i, j, new_count))
                new_old[new_sym] = A[i, j] * scale
                for old_sym in [A[i, j], A[j, j]]:
                    storage = old_new.get(old_sym)
                    if storage is None:
                        storage = set()
                        old_new[old_sym] = storage
                    storage.add(new_sym)

                # Change entry to the new symbol
                A[i, j] = new_sym
                new_count += 1

        nnz = 0
        for i in range(n):
            for j in range(n):
                nnz += not _iszero(A[i, j])

        total_op = new_count + zero_operations
        if total_op != 0:
            print(
                "Num non zeros in factorized jacobian:",
                nnz,
                int(nnz * 1.0 / n2 * 100),
                "%",
            )
            print(
                "Num non-zero operations while factorizing matrix:",
                new_count,
                new_count * 1.0 / total_op * 100,
                "%",
            )
            print(
                "Num zero operations while factorizing matrix:",
                zero_operations,
                int(zero_operations * 1.0 / total_op * 100),
                "%",
            )
        factorizing_nnz_operations = new_count

        self._jacobian_factorization_operations = operations
        self._factorized_jacobian = A

    def _compute_symbolic_fb_substitution(self):

        if hasattr(self, "_jacobian_fb_substitution_operations"):
            return

        if not hasattr(self, "_factorized_jacobian"):
            self._compute_symbolic_factorization_of_jacobian()

        global zero_operations
        zero_operations = 0

        A = self._factorized_jacobian

        n = A.rows
        b = sp.Matrix(n, 1, lambda i, j: sp.Symbol("F_{}".format(i)))

        old_new = OrderedDict()
        new_old = OrderedDict()

        # b = rhs.permuteFwd(p)

        operations = []
        new_count = 0

        def update_entry(new_count, i, j):
            global zero_operations
            if _iszero(b[j, 0] * A[i, j]):
                zero_operations += 1
                return new_count

            if i != 0:
                operations.append(
                    "dx[{i}] -= dx[{j}]*jac[{i}*{n}+{j}]".format(i=i, j=j, n=n)
                )
            else:
                operations.append("dx[{i}] -= dx[{j}]*jac[{j}]".format(i=i, j=j))

            # Create new symbol and store the representation
            new_sym = sp.Symbol("F_{0}:{1}".format(i, new_count))
            new_old[new_sym] = b[i, 0] - b[j, 0] * A[i, j]
            for old_sym in [b[i, 0], b[j, 0], A[i, j]]:
                storage = old_new.get(old_sym)
                if storage is None:
                    storage = set()
                    old_new[old_sym] = storage
                storage.add(new_sym)

            # Change entry to the new symbol
            b[i, 0] = new_sym
            new_count += 1

            return new_count

        # forward substitution, all diag entries are scaled to 1
        for i in range(n):
            for j in range(i):
                new_count = update_entry(new_count, i, j)
                # b.row(i, lambda x,k: x - b[j,k]*A[i,j])

        # backward substitution
        for i in range(n - 1, -1, -1):
            for j in range(i + 1, n):
                new_count = update_entry(new_count, i, j)
                # b.row(i, lambda x,k: x - b[j,k]*A[i,j])
            operations.append("dx[{i}] /= jac[{i}*{n}+{i}]".format(i=i, n=n))
            b.row(i, lambda x, k: x / A[i, i])

        total_count = zero_operations + new_count * 1.0
        if total_count != 0:
            print("Num operations while forward/backward substituting matrix:")
            print(
                "Zero operations:",
                zero_operations,
                int(100 * zero_operations / total_count),
                "%",
            )
            print(
                "Non-zero operations:",
                new_count,
                int(100 * new_count / total_count),
                "%",
            )

        self._jacobian_fb_substitution_operations = operations

    def _compute_linearized_dy_cse(self):
        self._compute_linearized_dy()

        if self._cse_subs_single_dy is not None:
            return

        info(
            "Calculating common sub expressions for single components of ODE. May take some time..."
        )
        sys.stdout.flush()
        ode = self.ode

        self._cse_subs_single_dy = []
        self._cse_derivative_expr_single_dy = []

        # Iterate over the derivative terms and collect information
        for i, (ders, expr) in enumerate(ode.get_derivative_expr(True)):
            cse_subs, cse_derivative_expr = cse(
                self.subs(expr),
                symbols=sp.numbered_symbols("cse_der_{0}_".format(i)),
                optimizations=[],
            )
            self._cse_subs_single_dy.append(cse_subs)
            self._cse_derivative_expr_single_dy.append(cse_derivative_expr)

        info(" done")
        info(
            "Calculating common sub expressions for linearized ODE. May take some time..."
        )
        self._cse_linearized_subs, self._cse_linearized_derivative_expr = cse(
            [self.subs(expr) for expr in list(self._linearized_exprs.values())],
            symbols=sp.numbered_symbols("cse_linear_"),
            optimizations=[],
        )

        info(" done")

    def _compute_linearized_dy(self):
        if self._linear_terms is not None:
            return

        ode = self.ode

        used_in_single_dy = []
        used_in_linear_dy = dict(parameters=set(), states=set())

        linearized_exprs = OrderedDict()

        # Iterate over the derivative terms and collect information
        for i, (ders, expr) in enumerate(ode.get_derivative_expr(True)):
            used_states = set()
            used_parameters = set()
            for sym in iter_symbol_params_from_expr(expr):
                if ode.has_state(sym):
                    used_states.add(sym.name)
                elif ode.has_parameter(sym):
                    used_parameters.add(sym.name)

            used_in_single_dy.append(
                dict(parameters=list(used_parameters), states=list(used_states))
            )

            assert len(ders) == 1

            # Grab state for the derivative
            state_sym = ders[0].sym
            expr_diff = expr.diff(state_sym)

            # Check for linear term
            if expr_diff and state_sym not in expr_diff.atoms():

                linearized_exprs[i] = expr_diff

                for sym in iter_symbol_params_from_expr(expr_diff):
                    if ode.has_state(sym):
                        used_in_linear_dy["states"].add(sym.name)
                    elif ode.has_parameter(sym):
                        used_in_linear_dy["parameters"].add(sym.name)

        # Store data
        self.linear_terms = [i in linearized_exprs for i in range(ode.num_states)]
        self._used_in_single_dy = used_in_single_dy
        self._linearized_exprs = linearized_exprs
        self._used_in_linear_dy = dict(
            states=list(used_in_linear_dy["states"]),
            parameters=list(used_in_linear_dy["parameters"]),
        )

    def update_index(self, index):
        """
        Set index notation, specific for language syntax
        """
        self.index = index

    @property
    def name(self):
        return self._name

    @property
    def class_name(self):
        name = self.name
        return (
            name
            if name[0].isupper()
            else name[0].upper() + (name[1:] if len(name) > 1 else "")
        )

    def set_state_prefix(self, prefix):
        """
        Register a prefix to a state name. Used if 
        """
        check_arg(prefix, str)
        self._state_prefix = prefix

        # Reset symbol subs
        self._symbol_subs = None

    def set_parameter_prefix(self, prefix):
        """
        Register a prefix to a parameter name. Used if 
        """
        check_arg(prefix, str)
        self._parameter_prefix = prefix

        # Reset symbol subs
        self._symbol_subs = None

    def subs(self, expr):
        """
        Call subs on the passed expr using symbol_subs if the expr is
        a SymPy Basic
        """
        if isinstance(expr, sp.Basic):
            return expr.subs(self.symbol_subs)
        return expr

    @property
    def symbol_subs(self):
        """
        Return a subs dict for all ODE Objects (states, parameters)
        """
        if self._symbol_subs is None:

            subs = []

            # Deal with parameter subs first
            if self.optimization.parameter_numerals:
                subs.extend((param.sym, param.init) for param in self.ode.parameters)
            elif not self.optimization.use_parameter_names:
                subs.extend(
                    (param.sym, sp.Symbol("parameters" + self.index(ind)))
                    for ind, param in enumerate(self.ode.parameters)
                )
            elif self._parameter_prefix:
                subs.extend(
                    (
                        param.sym,
                        sp.Symbol("{0}{1}".format(self._parameter_prefix, param.name)),
                    )
                    for param in self.ode.parameters
                )

            # Deal with state subs
            if not self.optimization.use_state_names:
                subs.extend(
                    (state.sym, sp.Symbol("states" + self.index(ind)))
                    for ind, state in enumerate(self.ode.states)
                )

            elif self._state_prefix:
                subs.extend(
                    (
                        param.sym,
                        sp.Symbol("{0}{1}".format(self._state_prefix, param.name)),
                    )
                    for param in self.ode.states
                )

            self._symbol_subs = subs

        return self._symbol_subs

    def iter_derivative_expr(self):
        """
        Return a list of derivatives and its expressions
        """

        # Keep intermediates is the lowest form for optimization deal with
        # first
        if self.optimization.keep_intermediates:

            return (
                (derivatives, self.subs(expr))
                for derivatives, expr in self.ode.get_derivative_expr()
            )

        # No intermediates and no CSE
        if not self.optimization.use_cse:
            return (
                (derivatives, self.subs(expr))
                for derivatives, expr in self.ode.get_derivative_expr(True)
            )

        # Use CSE
        else:
            self._compute_dy_cse()
            return (
                (derivatives, cse_expr)
                for ((derivatives, expr), cse_expr) in zip(
                    self.ode.get_derivative_expr(), self._cse_derivative_expr
                )
            )

    def iter_dy_body(self):
        """
        Return an interator over dy_body lines

        If using intermediates it will define these,
        if using cse extraction these will be returned
        """

        if self.optimization.keep_intermediates:

            # Iterate over the intermediates
            for intermediate in self.ode.intermediates:
                if isinstance(intermediate, (Comment, ODEComponent)):
                    yield intermediate.name, "COMMENT"
                    continue

                yield self.subs(intermediate.expr), intermediate.name

        elif self.optimization.use_cse:
            self._compute_dy_cse()
            yield "Common Sub Expressions", "COMMENT"
            for (name, expr), cse_count in zip(self._cse_subs, self._cse_counts):
                if cse_count:
                    yield expr, name

    def iter_jacobian_body(self):
        """
        Iterate over the body defining the jacobi expressions 
        """

        if not self.optimization.use_cse:
            return

        self._compute_jacobian_cse()

        # Yield the CSE
        yield "Common Sub Expressions for jacobi intermediates", "COMMENT"
        for (name, expr), cse_count in zip(
            self._cse_jacobian_subs, self._cse_jacobian_counts
        ):
            if cse_count:
                yield expr, name

    def iter_jacobian_expr(self):
        """
        Iterate over the jacobi expressions 
        """

        self._compute_jacobian_cse()

        if self.optimization.use_cse:

            for ((name, expr), cse_expr) in zip(
                list(self._jacobian_expr.items()), self._cse_jacobian_expr
            ):
                yield list(map(int, re.findall(_jacobian_pattern, str(name)))), cse_expr
        else:

            for name, expr in list(self._jacobian_expr.items()):
                yield list(
                    map(int, re.findall(_jacobian_pattern, str(name)))
                ), self.subs(expr)

    def iter_jacobian_action_body(self):
        """
        Iterate over the body defining the jacobi expressions 
        """

        if not self.optimization.use_cse:
            return

        self._compute_jacobian_action_cse()

        # Yield the CSE
        yield "Common Sub Expressions for jacobi action intermediates", "COMMENT"
        for (name, expr), cse_count in zip(
            self._cse_jacobian_action_subs, self._cse_jacobian_action_counts
        ):
            if cse_count:
                yield expr, name

    def iter_jacobian_action_expr(self):
        """
        Iterate over the jacobi expressions 
        """

        self._compute_jacobian_action_cse()

        if self.optimization.use_cse:

            for cse_expr in self._cse_jacobian_action_expr:
                yield cse_expr
        else:

            for expr in self._jacobian_action_expr:
                yield self.subs(expr)

    def iter_monitored_body(self):
        """
        Iterate over the body defining the monitored expressions 
        """

        self._compute_monitor_cse()
        if self.ode.num_monitored_intermediates == 0:
            return

        # Yield the CSE
        yield "Common Sub Expressions for monitored intermediates", "COMMENT"
        for (name, expr), cse_count in zip(
            self._cse_monitored_subs, self._cse_monitored_counts
        ):
            if cse_count:
                yield expr, name

    def iter_monitored_expr(self):
        """
        Iterate over the monitored expressions 
        """

        self._compute_monitor_cse()
        if self.ode.num_monitored_intermediates == 0:
            return

        yield "COMMENT", "Monitored intermediates"

        for (name, cse_expr) in zip(
            self.ode.monitored_intermediates, self._cse_monitored_expr
        ):
            yield name, cse_expr

    def iter_linerized_body(self):
        if not self.optimization.use_cse:
            return

        self._compute_linearized_dy_cse()

        # Yield the CSE
        yield "Common Sub Expressions for linearized evaluations", "COMMENT"
        for name, expr in self._cse_linearized_subs:
            yield expr, name

    def iter_linerized_expr(self):

        if self.optimization.use_cse:
            self._compute_linearized_dy_cse()
            for idx, cse_expr in zip(
                self._linearized_exprs, self._cse_linearized_derivative_expr
            ):
                yield idx, cse_expr

        else:
            self._compute_linearized_dy()
            for idx, expr in list(self._linearized_exprs.items()):
                yield idx, expr

    def iter_componentwise_dy(self):

        if self.optimization.use_cse:
            self._compute_linearized_dy_cse()
            for subs, expr in zip(
                self._cse_subs_single_dy, self._cse_derivative_expr_single_dy
            ):
                yield subs, expr[0]

        else:
            self._compute_linearized_dy()
            for ders, expr in self.ode.get_derivative_expr(True):
                yield [], expr
